# -*- coding: utf-8 -*-

from typing import Dict

import brainpy.math as bm
from .base import SDEIntegrator

__all__ = [
  'sdeint',
  'set_default_sdeint',
  'get_default_sdeint',
  'register_sde_integrator',
  'get_supported_methods',
]

name2method = {
}

_DEFAULT_SDE_METHOD = 'euler'


def sdeint(f=None,
           g=None,
           method='euler',
           dt: float = None,
           name: str = None,
           show_code: bool = False,
           var_type: str = None,
           intg_type: str = None,
           wiener_type: str = None,
           state_delays: Dict[str, bm.AbstractDelay] = None):
  """Numerical integration for SDEs.

  Parameters
  ----------
  f : callable, function
    The derivative function.
  method : str
    The shortcut name of the numerical integrator.

  Returns
  -------
  integral : SDEIntegrator
      The numerical solver of `f`.
  """
  method = _DEFAULT_SDE_METHOD if method is None else method
  if method not in name2method:
    raise ValueError(f'Unknown SDE numerical method "{method}". Currently '
                     f'BrainPy only support: {list(name2method.keys())}')

  if f is not None and g is not None:
    return name2method[method](f=f,
                               g=g,
                               dt=dt,
                               name=name,
                               show_code=show_code,
                               var_type=var_type,
                               intg_type=intg_type,
                               wiener_type=wiener_type,
                               state_delays=state_delays)

  elif f is not None:
    return lambda g: name2method[method](f=f,
                                         g=g,
                                         dt=dt,
                                         name=name,
                                         show_code=show_code,
                                         var_type=var_type,
                                         intg_type=intg_type,
                                         wiener_type=wiener_type,
                                         state_delays=state_delays)

  elif g is not None:
    return lambda f: name2method[method](f=f,
                                         g=g,
                                         dt=dt,
                                         name=name,
                                         show_code=show_code,
                                         var_type=var_type,
                                         intg_type=intg_type,
                                         wiener_type=wiener_type,
                                         state_delays=state_delays)

  else:
    raise ValueError('Must provide "f" or "g".')


def set_default_sdeint(method):
  """Set the default SDE numerical integrator method for differential equations.

  Parameters
  ----------
  method : str, callable
      Numerical integrator method.
  """
  if not isinstance(method, str):
    raise ValueError(f'Only support string, not {type(method)}.')
  if method not in name2method:
    raise ValueError(f'Unsupported SDE_INT numerical method: {method}.')

  global _DEFAULT_SDE_METHOD
  _DEFAULT_SDE_METHOD = method


def get_default_sdeint():
  """Get the default SDE numerical integrator method.

  Returns
  -------
  method : str
      The default numerical integrator method.
  """
  return _DEFAULT_SDE_METHOD


def register_sde_integrator(name, integrator):
  """Register a new SDE integrator.

  Parameters
  ----------
  name: ste
  integrator: type
  """
  if name in name2method:
    raise ValueError(f'"{name}" has been registered in SDE integrators.')
  if not issubclass(integrator, SDEIntegrator):
    raise ValueError(f'"integrator" must be an instance of {SDEIntegrator.__name__}')
  name2method[name] = integrator


def get_supported_methods():
  """Get all supported numerical methods for DDEs."""
  return list(name2method.keys())



# -*- coding: utf-8 -*-

from typing import Union, Sequence, Callable

import jax.numpy as jnp
from jax import jit, vmap
from jax import ops as jops
from jax.abstract_arrays import ShapedArray

from brainpy.errors import PackageMissingError, MathError
from brainpy.math import setting
from brainpy.math.jaxarray import JaxArray
from brainpy.math.numpy_ops import as_device_array, _remove_jaxarray
from brainpy.types import Shape

try:
  import brainpylib
except ModuleNotFoundError:
  brainpylib = None

__all__ = [
  # pre-to-post
  'pre2post_sum',
  'pre2post_prod',
  'pre2post_max',
  'pre2post_min',
  'pre2post_mean',

  # pre-to-syn
  'pre2syn',

  # syn-to-post
  'syn2post_sum', 'syn2post',
  'syn2post_prod',
  'syn2post_max',
  'syn2post_min',
  'syn2post_mean',
  'syn2post_softmax',

  # pre-to-post event operator
  'pre2post_event_sum',
  'pre2post_event_prod',

  # others
  'sparse_matmul',

  # numba operators
  'register_op'
]

_BRAINPYLIB_MINIMAL_VERSION = '0.0.4'

_pre2post = vmap(lambda pre_ids, pre_vs: pre_vs[pre_ids].sum(), in_axes=(0, None))
_pre2syn = vmap(lambda pre_id, pre_vs: pre_vs[pre_id], in_axes=(0, None))
_jit_seg_sum = jit(jops.segment_sum, static_argnums=(2, 3))
_jit_seg_prod = jit(jops.segment_prod, static_argnums=(2, 3))
_jit_seg_max = jit(jops.segment_max, static_argnums=(2, 3))
_jit_seg_min = jit(jops.segment_min, static_argnums=(2, 3))


def _check_brainpylib(ops_name):
  if brainpylib is not None:
    if brainpylib.__version__ < _BRAINPYLIB_MINIMAL_VERSION:
      raise PackageMissingError(
        f'"{ops_name}" operator need "brainpylib>={_BRAINPYLIB_MINIMAL_VERSION}". \n'
        f'Please install it through:\n\n'
        f'>>> pip install brainpylib>={_BRAINPYLIB_MINIMAL_VERSION} -U'
      )
  else:
    raise PackageMissingError(
      f'"brainpylib" must be installed when the user '
      f'wants to use "{ops_name}" operator. \n'
      f'Please install "brainpylib>={_BRAINPYLIB_MINIMAL_VERSION}" through:\n\n'
      f'>>> pip install brainpylib>={_BRAINPYLIB_MINIMAL_VERSION}'
    )


def register_op(
    op_name: str,
    cpu_func: Callable,
    gpu_func: Callable = None,
    out_shapes: Union[Callable, ShapedArray, Sequence[ShapedArray]] = None,
    apply_cpu_func_to_gpu: bool = False
):
  """
  Converting the numba-jitted function in a Jax/XLA compatible primitive.

  Parameters
  ----------
  op_name: str
    Name of the operators.
  cpu_func: Callble
    A callable numba-jitted function or pure function (can be lambda function) running on CPU.
  gpu_func: Callable, default = None
    A callable cuda-jitted kernel running on GPU.
  out_shapes: Callable, ShapedArray, Sequence[ShapedArray], default = None
    Outputs shapes of target function. `out_shapes` can be a `ShapedArray` or
    a sequence of `ShapedArray`. If it is a function, it takes as input the argument
    shapes and dtypes and should return correct output shapes of `ShapedArray`.
  apply_cpu_func_to_gpu: bool, default = False
    True when gpu_func is implemented on CPU and other logics(data transfer) is implemented on GPU.

  Returns
  -------
  A jitable JAX function.
  """
  _check_brainpylib(register_op.__name__)
  f = brainpylib.register_op(op_name, cpu_func, gpu_func, out_shapes, apply_cpu_func_to_gpu)

  def fixed_op(*inputs):
    inputs = tuple([i.value if isinstance(i, JaxArray) else i for i in inputs])
    return f(*inputs)

  return fixed_op


def pre2post_event_sum(events, pre2post, post_num, values=1.):
  """The pre-to-post synaptic computation with event-driven summation.

  When ``values`` is a scalar, this function is equivalent to

  .. highlight:: python
  .. code-block:: python

    post_val = np.zeros(post_num)
    post_ids, idnptr = pre2post
    for i in range(pre_num):
      if events[i]:
        for j in range(idnptr[i], idnptr[i+1]):
          post_val[post_ids[i]] += values

  When ``values`` is a vector (with the length of ``len(post_ids)``),
  this function is equivalent to

  .. highlight:: python
  .. code-block:: python

    post_val = np.zeros(post_num)

    post_ids, idnptr = pre2post
    for i in range(pre_num):
      if events[i]:
        for j in range(idnptr[i], idnptr[i+1]):
          post_val[post_ids[i]] += values[j]


  Parameters
  ----------
  events: JaxArray, jax.numpy.ndarray, Variable
    The events, must be bool.
  pre2post: tuple of JaxArray, tuple of jax.numpy.ndarray
    A tuple contains the connection information of pre-to-post.
  post_num: int
    The number of post-synaptic group.
  values: float, JaxArray, jax.numpy.ndarray
    The value to make summation.

  Returns
  -------
  out: JaxArray, jax.numpy.ndarray
    A tensor with the shape of ``post_num``.
  """
  _check_brainpylib(pre2post_event_sum.__name__)
  indices, idnptr = pre2post
  events = as_device_array(events)
  indices = as_device_array(indices)
  idnptr = as_device_array(idnptr)
  values = as_device_array(values)
  return brainpylib.event_sum(events, (indices, idnptr), post_num, values)


def pre2post_event_prod(events, pre2post, post_num, values=1.):
  """The pre-to-post synaptic computation with event-driven production.

  When ``values`` is a scalar, this function is equivalent to

  .. highlight:: python
  .. code-block:: python

    post_val = np.ones(post_num)
    post_ids, idnptr = pre2post
    for i in range(pre_num):
      if events[i]:
        for j in range(idnptr[i], idnptr[i+1]):
          post_val[post_ids[i]] *= values

  When ``values`` is a vector (with the length of ``len(post_ids)``),
  this function is equivalent to

  .. highlight:: python
  .. code-block:: python

    post_val = np.ones(post_num)

    post_ids, idnptr = pre2post
    for i in range(pre_num):
      if events[i]:
        for j in range(idnptr[i], idnptr[i+1]):
          post_val[post_ids[i]] *= values[j]


  Parameters
  ----------
  events: JaxArray, jax.numpy.ndarray, Variable
    The events, must be bool.
  pre2post: tuple of JaxArray, tuple of jax.numpy.ndarray
    A tuple contains the connection information of pre-to-post.
  post_num: int
    The number of post-synaptic group.
  values: float, JaxArray, jax.numpy.ndarray
    The value to make summation.

  Returns
  -------
  out: JaxArray, jax.numpy.ndarray
    A tensor with the shape of ``post_num``.
  """
  _check_brainpylib(pre2post_event_prod.__name__)
  indices, idnptr = pre2post
  events = as_device_array(events)
  indices = as_device_array(indices)
  idnptr = as_device_array(idnptr)
  values = as_device_array(values)
  return brainpylib.event_prod(events, (indices, idnptr), post_num, values)


def _raise_pre_ids_is_none(pre_ids):
  if pre_ids is None:
    raise MathError(f'pre2post synaptic computation needs "pre_ids" '
                    f'when providing heterogeneous "pre_values" '
                    f'(brainpy.math.ndim(pre_values) != 0).')


def pre2post_sum(pre_values, post_num, post_ids, pre_ids=None):
  """The pre-to-post synaptic summation.

  This function is equivalent to:

  .. highlight:: python
  .. code-block:: python

     post_val = np.zeros(post_num)
     for i, j in zip(pre_ids, post_ids):
       post_val[j] += pre_values[pre_ids[i]]

  Parameters
  ----------
  pre_values: float, jax.numpy.ndarray, JaxArray, Variable
    The pre-synaptic values.
  post_ids: jax.numpy.ndarray, JaxArray
    The connected post-synaptic neuron ids.
  post_num: int
    Output dimension. The number of post-synaptic neurons.
  pre_ids: optional, jax.numpy.ndarray, JaxArray
    The connected pre-synaptic neuron ids.

  Returns
  -------
  post_val: jax.numpy.ndarray, JaxArray
    The value with the size of post-synaptic neurons.
  """
  out = jnp.zeros(post_num, dtype=setting.float_)
  pre_values = as_device_array(pre_values)
  post_ids = as_device_array(post_ids)
  if jnp.ndim(pre_values) != 0:
    _raise_pre_ids_is_none(pre_ids)
    pre_ids = as_device_array(pre_ids)
    pre_values = pre_values[pre_ids]
  return out.at[post_ids].add(pre_values)


def pre2post_prod(pre_values, post_num, post_ids, pre_ids=None):
  """The pre-to-post synaptic production.

  This function is equivalent to:

  .. highlight:: python
  .. code-block:: python

     post_val = np.zeros(post_num)
     for i, j in zip(pre_ids, post_ids):
       post_val[j] *= pre_values[pre_ids[i]]

  Parameters
  ----------
  pre_values: float, jax.numpy.ndarray, JaxArray, Variable
    The pre-synaptic values.
  pre_ids: jax.numpy.ndarray, JaxArray
    The connected pre-synaptic neuron ids.
  post_ids: jax.numpy.ndarray, JaxArray
    The connected post-synaptic neuron ids.
  post_num: int
    Output dimension. The number of post-synaptic neurons.

  Returns
  -------
  post_val: jax.numpy.ndarray, JaxArray
    The value with the size of post-synaptic neurons.
  """
  out = jnp.zeros(post_num, dtype=setting.float_)
  pre_values = as_device_array(pre_values)
  post_ids = as_device_array(post_ids)
  if jnp.ndim(pre_values) != 0:
    _raise_pre_ids_is_none(pre_ids)
    pre_ids = as_device_array(pre_ids)
    pre_values = pre_values[pre_ids]
  return out.at[post_ids].multiply(pre_values)


def pre2post_min(pre_values, post_num, post_ids, pre_ids=None):
  """The pre-to-post synaptic minimization.

  This function is equivalent to:

  .. highlight:: python
  .. code-block:: python

     post_val = np.zeros(post_num)
     for i, j in zip(pre_ids, post_ids):
       post_val[j] = np.minimum(post_val[j], pre_values[pre_ids[i]])

  Parameters
  ----------
  pre_values: float, jax.numpy.ndarray, JaxArray
    The pre-synaptic values.
  pre_ids: jax.numpy.ndarray, JaxArray
    The connected pre-synaptic neuron ids.
  post_ids: jax.numpy.ndarray, JaxArray
    The connected post-synaptic neuron ids.
  post_num: int
    Output dimension. The number of post-synaptic neurons.

  Returns
  -------
  post_val: jax.numpy.ndarray, JaxArray
    The value with the size of post-synaptic neurons.
  """
  out = jnp.zeros(post_num, dtype=setting.float_)
  pre_values = as_device_array(pre_values)
  post_ids = as_device_array(post_ids)
  if jnp.ndim(pre_values) != 0:
    _raise_pre_ids_is_none(pre_ids)
    pre_ids = as_device_array(pre_ids)
    pre_values = pre_values[pre_ids]
  return out.at[post_ids].min(pre_values)


def pre2post_max(pre_values, post_num, post_ids, pre_ids=None):
  """The pre-to-post synaptic maximization.

  This function is equivalent to:

  .. highlight:: python
  .. code-block:: python

     post_val = np.zeros(post_num)
     for i, j in zip(pre_ids, post_ids):
       post_val[j] = np.maximum(post_val[j], pre_values[pre_ids[i]])

  Parameters
  ----------
  pre_values: float, jax.numpy.ndarray, JaxArray, Variable
    The pre-synaptic values.
  pre_ids: jax.numpy.ndarray, JaxArray
    The connected pre-synaptic neuron ids.
  post_ids: jax.numpy.ndarray, JaxArray
    The connected post-synaptic neuron ids.
  post_num: int
    Output dimension. The number of post-synaptic neurons.

  Returns
  -------
  post_val: jax.numpy.ndarray, JaxArray
    The value with the size of post-synaptic neurons.
  """
  out = jnp.zeros(post_num, dtype=setting.float_)
  pre_values = as_device_array(pre_values)
  post_ids = as_device_array(post_ids)
  if jnp.ndim(pre_values) != 0:
    _raise_pre_ids_is_none(pre_ids)
    pre_ids = as_device_array(pre_ids)
    pre_values = pre_values[pre_ids]
  return out.at[post_ids].max(pre_values)


def pre2post_mean(pre_values, post_num, post_ids, pre_ids=None):
  """The pre-to-post synaptic mean computation.

  Parameters
  ----------
  pre_values: float, jax.numpy.ndarray, JaxArray, Variable
    The pre-synaptic values.
  pre_ids: jax.numpy.ndarray, JaxArray
    The connected pre-synaptic neuron ids.
  post_ids: jax.numpy.ndarray, JaxArray
    The connected post-synaptic neuron ids.
  post_num: int
    Output dimension. The number of post-synaptic neurons.

  Returns
  -------
  post_val: jax.numpy.ndarray, JaxArray
    The value with the size of post-synaptic neurons.
  """
  out = jnp.zeros(post_num, dtype=setting.float_)
  pre_values = as_device_array(pre_values)
  post_ids = as_device_array(post_ids)
  if jnp.ndim(pre_values) == 0:
    # return out.at[post_ids].set(pre_values)
    return out.at[jnp.unique(post_ids)].set(pre_values)
  else:
    _raise_pre_ids_is_none(pre_ids)
    pre_ids = as_device_array(pre_ids)
    pre_values = pre2syn(pre_values, pre_ids)
    return syn2post_mean(pre_values, post_ids, post_num)


def pre2syn(pre_values, pre_ids):
  """The pre-to-syn computation.

  Change the pre-synaptic data to the data with the dimension of synapses.

  This function is equivalent to:

  .. highlight:: python
  .. code-block:: python

    syn_val = np.zeros(len(pre_ids))
    for syn_i, pre_i in enumerate(pre_ids):
      syn_val[i] = pre_values[pre_i]

  Parameters
  ----------
  pre_values: float, jax.numpy.ndarray, JaxArray, Variable
    The pre-synaptic value.
  pre_ids: jax.numpy.ndarray, JaxArray
    The pre-synaptic neuron index.

  Returns
  -------
  syn_val: jax.numpy.ndarray, JaxArray
    The synaptic value.
  """
  pre_values = as_device_array(pre_values)
  pre_ids = as_device_array(pre_ids)
  if jnp.ndim(pre_values) == 0:
    return jnp.ones(len(pre_ids), dtype=pre_values.dtype) * pre_values
  else:
    return _pre2syn(pre_ids, pre_values)


def syn2post_sum(syn_values, post_ids, post_num: int, indices_are_sorted=True):
  """The syn-to-post summation computation.

  This function is equivalent to:

  .. highlight:: python
  .. code-block:: python

    post_val = np.zeros(post_num)
    for syn_i, post_i in enumerate(post_ids):
      post_val[post_i] += syn_values[syn_i]

  Parameters
  ----------
  syn_values: jax.numpy.ndarray, JaxArray, Variable
    The synaptic values.
  post_ids: jax.numpy.ndarray, JaxArray
    The post-synaptic neuron ids.
  post_num: int
    The number of the post-synaptic neurons.

  Returns
  -------
  post_val: jax.numpy.ndarray, JaxArray
    The post-synaptic value.
  """
  post_ids = as_device_array(post_ids)
  syn_values = as_device_array(syn_values)
  if syn_values.dtype == jnp.bool_:
    syn_values = jnp.asarray(syn_values, dtype=jnp.int32)
  return _jit_seg_sum(syn_values, post_ids, post_num, indices_are_sorted)


syn2post = syn2post_sum


def syn2post_prod(syn_values, post_ids, post_num: int, indices_are_sorted=True):
  """The syn-to-post product computation.

  This function is equivalent to:

  .. highlight:: python
  .. code-block:: python

    post_val = np.zeros(post_num)
    for syn_i, post_i in enumerate(post_ids):
      post_val[post_i] *= syn_values[syn_i]

  Parameters
  ----------
  syn_values: jax.numpy.ndarray, JaxArray, Variable
    The synaptic values.
  post_ids: jax.numpy.ndarray, JaxArray
    The post-synaptic neuron ids. If ``post_ids`` is generated by
    ``brainpy.conn.TwoEndConnector``, then it has sorted indices.
    Otherwise, this function cannot guarantee indices are sorted.
    You's better set ``indices_are_sorted=False``.
  post_num: int
    The number of the post-synaptic neurons.
  indices_are_sorted: whether ``post_ids`` is known to be sorted.

  Returns
  -------
  post_val: jax.numpy.ndarray, JaxArray
    The post-synaptic value.
  """
  post_ids = as_device_array(post_ids)
  syn_values = as_device_array(syn_values)
  if syn_values.dtype == jnp.bool_:
    syn_values = jnp.asarray(syn_values, dtype=jnp.int32)
  return _jit_seg_prod(syn_values, post_ids, post_num, indices_are_sorted)


def syn2post_max(syn_values, post_ids, post_num: int, indices_are_sorted=True):
  """The syn-to-post maximum computation.

  This function is equivalent to:

  .. highlight:: python
  .. code-block:: python

    post_val = np.zeros(post_num)
    for syn_i, post_i in enumerate(post_ids):
      post_val[post_i] = np.maximum(post_val[post_i], syn_values[syn_i])

  Parameters
  ----------
  syn_values: jax.numpy.ndarray, JaxArray, Variable
    The synaptic values.
  post_ids: jax.numpy.ndarray, JaxArray
    The post-synaptic neuron ids. If ``post_ids`` is generated by
    ``brainpy.conn.TwoEndConnector``, then it has sorted indices.
    Otherwise, this function cannot guarantee indices are sorted.
    You's better set ``indices_are_sorted=False``.
  post_num: int
    The number of the post-synaptic neurons.
  indices_are_sorted: whether ``post_ids`` is known to be sorted.

  Returns
  -------
  post_val: jax.numpy.ndarray, JaxArray
    The post-synaptic value.
  """
  post_ids = as_device_array(post_ids)
  syn_values = as_device_array(syn_values)
  if syn_values.dtype == jnp.bool_:
    syn_values = jnp.asarray(syn_values, dtype=jnp.int32)
  return _jit_seg_max(syn_values, post_ids, post_num, indices_are_sorted)


def syn2post_min(syn_values, post_ids, post_num: int, indices_are_sorted=True):
  """The syn-to-post minimization computation.

  This function is equivalent to:

  .. highlight:: python
  .. code-block:: python

    post_val = np.zeros(post_num)
    for syn_i, post_i in enumerate(post_ids):
      post_val[post_i] = np.minimum(post_val[post_i], syn_values[syn_i])

  Parameters
  ----------
  syn_values: jax.numpy.ndarray, JaxArray, Variable
    The synaptic values.
  post_ids: jax.numpy.ndarray, JaxArray
    The post-synaptic neuron ids. If ``post_ids`` is generated by
    ``brainpy.conn.TwoEndConnector``, then it has sorted indices.
    Otherwise, this function cannot guarantee indices are sorted.
    You's better set ``indices_are_sorted=False``.
  post_num: int
    The number of the post-synaptic neurons.
  indices_are_sorted: whether ``post_ids`` is known to be sorted.

  Returns
  -------
  post_val: jax.numpy.ndarray, JaxArray
    The post-synaptic value.
  """
  post_ids = as_device_array(post_ids)
  syn_values = as_device_array(syn_values)
  if syn_values.dtype == jnp.bool_:
    syn_values = jnp.asarray(syn_values, dtype=jnp.int32)
  return _jit_seg_min(syn_values, post_ids, post_num, indices_are_sorted)


def syn2post_mean(syn_values, post_ids, post_num: int, indices_are_sorted=True):
  """The syn-to-post mean computation.

  Parameters
  ----------
  syn_values: jax.numpy.ndarray, JaxArray, Variable
    The synaptic values.
  post_ids: jax.numpy.ndarray, JaxArray
    The post-synaptic neuron ids. If ``post_ids`` is generated by
    ``brainpy.conn.TwoEndConnector``, then it has sorted indices.
    Otherwise, this function cannot guarantee indices are sorted.
    You's better set ``indices_are_sorted=False``.
  post_num: int
    The number of the post-synaptic neurons.
  indices_are_sorted: whether ``post_ids`` is known to be sorted.

  Returns
  -------
  post_val: jax.numpy.ndarray, JaxArray
    The post-synaptic value.
  """
  post_ids = as_device_array(post_ids)
  syn_values = as_device_array(syn_values)
  if syn_values.dtype == jnp.bool_:
    syn_values = jnp.asarray(syn_values, dtype=jnp.int32)
  nominator = _jit_seg_sum(syn_values, post_ids, post_num, indices_are_sorted)
  denominator = _jit_seg_sum(jnp.ones_like(syn_values), post_ids, post_num, indices_are_sorted)
  return jnp.nan_to_num(nominator / denominator)


def syn2post_softmax(syn_values, post_ids, post_num: int, indices_are_sorted=True):
  """The syn-to-post softmax computation.

  Parameters
  ----------
  syn_values: jax.numpy.ndarray, JaxArray, Variable
    The synaptic values.
  post_ids: jax.numpy.ndarray, JaxArray
    The post-synaptic neuron ids. If ``post_ids`` is generated by
    ``brainpy.conn.TwoEndConnector``, then it has sorted indices.
    Otherwise, this function cannot guarantee indices are sorted.
    You's better set ``indices_are_sorted=False``.
  post_num: int
    The number of the post-synaptic neurons.
  indices_are_sorted: whether ``post_ids`` is known to be sorted.

  Returns
  -------
  post_val: jax.numpy.ndarray, JaxArray
    The post-synaptic value.
  """
  post_ids = as_device_array(post_ids)
  syn_values = as_device_array(syn_values)
  if syn_values.dtype == jnp.bool_:
    syn_values = jnp.asarray(syn_values, dtype=jnp.int32)
  syn_maxs = _jit_seg_max(syn_values, post_ids, post_num, indices_are_sorted)
  syn_values = syn_values - syn_maxs[post_ids]
  syn_values = jnp.exp(syn_values)
  normalizers = _jit_seg_sum(syn_values, post_ids, post_num, indices_are_sorted)
  softmax = syn_values / normalizers[post_ids]
  return jnp.nan_to_num(softmax)


def _matmul_with_left_sparse(sparse: Sequence,
                             dense: Union[JaxArray, jnp.ndarray],
                             shape: int):
  r"""Matrix multiplication with sparse matrix on the left.

  .. math::

    Y = M_{\mathrm{sparse}} @ M_{\mathrm{dense}}

  Parameters
  ----------
  sparse: tuple, list
    The sparse matrix with shape of :math:`(N, M)`.
  dense: JaxArray, jnp.ndarray
    The dense matrix with the shape of :math:`(M, K)`.
  shape: int
    The dimension of :math:`N`.

  Returns
  -------
  matrix
    A tensor the the shape of :math:`(N, K)`.
  """
  assert dense.ndim in [1, 2], 'Dense matrix must be a one- or two-dimensional matrix.'
  values, (rows, cols) = sparse
  values = _remove_jaxarray(values)
  rows = _remove_jaxarray(rows)
  cols = _remove_jaxarray(cols)
  dense = _remove_jaxarray(dense)
  B = dense.take(cols, axis=0)
  if B.ndim == 2:
    prod = B * jnp.reshape(values, (-1, 1))
  else:
    prod = B * values
  return jops.segment_sum(prod, rows, shape)


def _matmul_with_right_sparse(dense, sparse, shape):
  r"""Matrix multiplication with sparse matrix on the left.

  .. math::

    Y = M_{\mathrm{dense}} @ M_{\mathrm{sparse}}

  Parameters
  ----------
  dense: JaxArray, jnp.ndarray
    The dense matrix with the shape of :math:`(N, M)`.
  sparse: tuple, list
    The sparse matrix with shape of :math:`(M, K)`.
  shape: int
    The dimension of :math:`K`.

  Returns
  -------
  matrix
    A tensor the the shape of :math:`(N, K)`.
  """
  assert dense.ndim in [1, 2], 'Dense matrix must be a one- or two-dimensional matrix.'
  values, (rows, cols) = sparse
  values = _remove_jaxarray(values)
  rows = _remove_jaxarray(rows)
  cols = _remove_jaxarray(cols)
  dense = _remove_jaxarray(dense)
  if dense.ndim == 2:
    A = dense[:, rows]
    prod = (A * values).T
    res = jops.segment_sum(prod, cols, shape).T
  else:
    prod = dense[rows] * values
    res = jops.segment_sum(prod, cols, shape)
  return res


def sparse_matmul(A, B, shape: Shape):
  r"""Sparse matrix multiplication.

  .. math::

     y = A @ B

  where :math:`A` or :math:`B` is a sparse matrix.
  :math:`A` and :math:`B` cannot be both sparse.

  Examples
  --------

  >>> import brainpy.math as bm

  1. when the left matrix :math:`A` is a sparse matrix with the shape of :math:`(N, M)`,
     we should provide :math:`N` as the ``shape`` in the ``brainpy.math.sparse_matmul``
     function.

  >>> # A is a sparse matrix (3, 4):
  >>> #   [[0, 2, 0, 4],
  >>> #    [1, 0, 0, 0],
  >>> #    [0, 3, 0, 2]]
  >>> values = bm.asarray([2, 4, 1, 3, 2])
  >>> rows = bm.asarray([0, 0, 1, 2, 2])
  >>> cols = bm.asarray([1, 3, 0, 1, 3])
  >>> B = bm.arange(4)
  >>> bm.sparse_matmul([values, (rows, cols)], B, 3)
  JaxArray([14,  0,  9], dtype=int32)
  >>> B = bm.random.rand(4, 3)
  >>> bm.sparse_matmul([values, (rows, cols)], B, 3)
  JaxArray([[3.8331761 , 1.3708692 , 4.510223  ],
            [0.9960836 , 0.37550318, 0.7370341 ],
            [2.3700516 , 0.7574289 , 4.1124535 ]], dtype=float32)

  2. when the right matrix :math:`B` is a sparse matrix with the shape of :math:`(M, K)`,
     we should provide :math:`K` as the ``shape`` in the ``brainpy.math.sparse_matmul``
     function.

  >>> A = bm.arange(3)
  >>> bm.sparse_matmul(A, [values, (rows, cols)], 4)
  JaxArray([1, 6, 0, 4], dtype=int32)
  >>> A = bm.random.rand(2, 3)
  JaxArray([[0.438388  , 1.4346815 , 0.        , 2.361964  ],
            [0.9171978 , 1.1214957 , 0.        , 0.90534496]],  dtype=float32)

  Parameters
  ----------
  A: tensor, sequence
    The dense or sparse matrix with the shape of :math:`(N, M)`.
  B: tensor, sequence
    The dense or sparse matrix with the shape of :math:`(M, K)`.
  shape: int
    The dimension of :math:`N` when ``A`` is sparse, or
    the dimension of :math:`K` when ``B`` is sparse.

  Returns
  -------
  results: JaxArray, jnp.ndarray
    The tensor with the shape of :math:`(N, K)`.
  """
  if isinstance(A, (tuple, list)):
    assert isinstance(B, (JaxArray, jnp.ndarray)), ('A and B cannot be both sparse. \n'
                                                    f'A:\n{A}\n'
                                                    f'B:\n{B}')
    return _matmul_with_left_sparse(A, B, shape)
  else:
    assert isinstance(B, (tuple, list)), ('A and B cannot be both dense. \n'
                                          f'A:\n{A}\n'
                                          f'B:\n{B}')
    return _matmul_with_right_sparse(A, B, shape)



# -*- coding: utf-8 -*-

import brainpy.math as bm
from brainpy.base import Base

__all__ = [
  # base class
  'OnlineAlgorithm',

  # online learning algorithms
  'ForceLearning',
  'RLS',
  'LMS',

  # generic methods
  'get_supported_online_methods',
  'register_online_method',
]

name2func = dict()


class OnlineAlgorithm(Base):
  """Base class for online training algorithm."""

  def __init__(self, name=None):
    super(OnlineAlgorithm, self).__init__(name=name)

  def __call__(self, name, target, input, output):
    """The training procedure.

    Parameters
    ----------
    name: str
      The variable name.
    target: JaxArray, ndarray
      The 2d target data with the shape of `(num_batch, num_output)`.
    input: JaxArray, ndarray
      The 2d input data with the shape of `(num_batch, num_input)`.
    output: JaxArray, ndarray
      The 2d output data with the shape of `(num_batch, num_output)`.

    Returns
    -------
    weight: JaxArray
      The weights after fit.
    """
    return self.call(name, target, input, output)

  def initialize(self, name, *args, **kwargs):
    raise NotImplementedError('Must implement the initialize() '
                              'function by the subclass itself.')

  def call(self, name, target, input, output):
    """The training procedure.

    Parameters
    ----------
    name: str
      The variable name.
    target: JaxArray, ndarray
      The 2d target data with the shape of `(num_batch, num_output)`.
    input: JaxArray, ndarray
      The 2d input data with the shape of `(num_batch, num_input)`.
    output: JaxArray, ndarray
      The 2d output data with the shape of `(num_batch, num_output)`.

    Returns
    -------
    weight: JaxArray
      The weights after fit.
    """
    raise NotImplementedError('Must implement the call() function by the subclass itself.')

  def __repr__(self):
    return self.__class__.__name__


class RLS(OnlineAlgorithm):
  """The recursive least squares (RLS)."""

  postfix = '.rls.P'

  def __init__(self, alpha=0.1, name=None):
    super(RLS, self).__init__(name=name)
    self.alpha = alpha

  def initialize(self, name, feature_in, feature_out=None):
    name = name + self.postfix
    self.implicit_vars[name] = bm.Variable(bm.eye(feature_in) * self.alpha)

  def call(self, name, target, input, output):
    name = name + self.postfix
    P = self.implicit_vars[name]
    # update the inverse correlation matrix
    k = bm.dot(P, input.T)  # (num_input, num_batch)
    hPh = bm.dot(input, k)  # (num_batch, num_batch)
    c = bm.sum(1.0 / (1.0 + hPh))  # ()
    P -= c * bm.dot(k, k.T)  # (num_input, num_input)
    # update weights
    e = output - target  # (num_batch, num_output)
    dw = -c * bm.dot(k, e)  # (num_input, num_output)
    return dw


name2func['rls'] = RLS


class ForceLearning(RLS):
  postfix = '.force.P'


name2func['force'] = ForceLearning


class LMS(OnlineAlgorithm):
  """The least mean squares (LMS). """

  def __init__(self, alpha=0.1, name=None):
    super(LMS, self).__init__(name=name)
    self.alpha = alpha

  def initialize(self, name, *args, **kwargs):
    pass

  def call(self, name, target, input, output):
    return -self.alpha * bm.dot(output - target, output)


name2func['lms'] = LMS


def get_supported_online_methods():
  """Get all supported online training methods."""
  return tuple(name2func.keys())


def register_online_method(name, method):
  """Register a new oneline learning method.

  Parameters
  ----------
  name: str
    The method name.
  method: callable
    The function method.
  """
  if name in name2func:
    raise ValueError(f'"{name}" has been registered in offline training methods.')
  if not callable(method):
    raise ValueError(f'"method" must be an instance of callable '
                     f'function, but we got {type(method)}')
  name2func[name] = method


def get(name):
  """Get the training function according to the training method name."""
  if name not in name2func:
    raise ValueError(f'All online methods are: {get_supported_online_methods()}.\n'
                     f'But we got {name}.')
  return name2func[name]



# -*- coding: utf-8 -*-

import inspect

from brainpy import errors, math
from brainpy.integrators import constants, utils
from brainpy.integrators.analysis_by_ast import separate_variables
from brainpy.integrators.sde.base import SDEIntegrator
from .generic import register_sde_integrator

try:
  import sympy
  from brainpy.integrators import analysis_by_sympy
except (ModuleNotFoundError, ImportError):
  sympy = analysis_by_sympy = None

__all__ = [
  'Euler',
  'Heun',
  'Milstein',
  'ExponentialEuler',
]


def df_and_dg(code_lines, variables, parameters):
  # 1. df
  # df = f(x, t, *args)
  all_df = [f'{var}_df' for var in variables]
  code_lines.append(f'  {", ".join(all_df)} = f({", ".join(variables + parameters)})')

  # 2. dg
  # dg = g(x, t, *args)
  all_dg = [f'{var}_dg' for var in variables]
  code_lines.append(f'  {", ".join(all_dg)} = g({", ".join(variables + parameters)})')
  code_lines.append('  ')


def dfdt(code_lines, variables):
  for var in variables:
    code_lines.append(f'  {var}_dfdt = {var}_df * {constants.DT}')
  code_lines.append('  ')


def noise_terms(code_lines, variables):
  # num_vars = len(variables)
  # if num_vars > 1:
  #     code_lines.append(f'  all_dW = math.normal(0.0, dt_sqrt, ({num_vars},)+math.shape({variables[0]}_dg))')
  #     for i, var in enumerate(variables):
  #         code_lines.append(f'  {var}_dW = all_dW[{i}]')
  # else:
  #     var = variables[0]
  #     code_lines.append(f'  {var}_dW = math.normal(0.0, dt_sqrt, math.shape({var}))')
  # code_lines.append('  ')

  for var in variables:
    code_lines.append(f'  {var}_dW = random.normal(0.000, dt_sqrt, math.shape({var})).value')
  code_lines.append('  ')


class Euler(SDEIntegrator):
  def __init__(self, f, g, dt=None, name=None, show_code=False,
               var_type=None, intg_type=None, wiener_type=None,
               state_delays=None):
    super(Euler, self).__init__(f=f, g=g, dt=dt, show_code=show_code, name=name,
                                var_type=var_type, intg_type=intg_type,
                                wiener_type=wiener_type, state_delays=state_delays)
    self.build()

  def build(self):
    self.code_lines.append(f'  {constants.DT}_sqrt = {constants.DT} ** 0.5')

    # 2.1 df, dg
    df_and_dg(self.code_lines, self.variables, self.parameters)

    # 2.2 dfdt
    dfdt(self.code_lines, self.variables)

    # 2.3 dW
    noise_terms(self.code_lines, self.variables)

    # 2.3 dgdW
    # ----
    # SCALAR_WIENER : dg * dW
    # VECTOR_WIENER : math.sum(dg * dW, axis=-1)

    if self.wiener_type == constants.SCALAR_WIENER:
      for var in self.variables:
        self.code_lines.append(f'  {var}_dgdW = {var}_dg * {var}_dW')
    else:
      for var in self.variables:
        self.code_lines.append(f'  {var}_dgdW = math.sum({var}_dg * {var}_dW, axis=-1)')
    self.code_lines.append('  ')

    if self.intg_type == constants.ITO_SDE:
      # 2.4 new var
      # ----
      # y = x + dfdt + dgdW
      for var in self.variables:
        self.code_lines.append(f'  {var}_new = {var} + {var}_dfdt + {var}_dgdW')
      self.code_lines.append('  ')

    elif self.intg_type == constants.STRA_SDE:
      # 2.4  y_bar = x + math.sum(dgdW, axis=-1)
      all_bar = [f'{var}_bar' for var in self.variables]
      for var in self.variables:
        self.code_lines.append(f'  {var}_bar = {var} + {var}_dgdW')
      self.code_lines.append('  ')

      # 2.5  dg_bar = g(y_bar, t, *args)
      all_dg_bar = [f'{var}_dg_bar' for var in self.variables]
      self.code_lines.append(f'  {", ".join(all_dg_bar)} = g({", ".join(all_bar + self.parameters)})')

      # 2.6 dgdW2
      # ----
      # SCALAR_WIENER : dgdW2 = dg_bar * dW
      # VECTOR_WIENER : dgdW2 = math.sum(dg_bar * dW, axis=-1)
      if self.wiener_type == constants.SCALAR_WIENER:
        for var in self.variables:
          self.code_lines.append(f'  {var}_dgdW2 = {var}_dg_bar * {var}_dW')
      else:
        for var in self.variables:
          self.code_lines.append(f'  {var}_dgdW2 = math.sum({var}_dg_bar * {var}_dW, axis=-1)')
      self.code_lines.append('  ')

      # 2.7 new var
      # ----
      # y = x + dfdt + 0.5 * (dgdW + dgdW2)
      for var in self.variables:
        self.code_lines.append(f'  {var}_new = {var} + {var}_dfdt + 0.5 * ({var}_dgdW + {var}_dgdW2)')
      self.code_lines.append('  ')
    else:
      raise ValueError(f'Unknown SDE_INT type: {self.intg_type}. We only '
                       f'supports {constants.SUPPORTED_INTG_TYPE}.')

    # returns
    new_vars = [f'{var}_new' for var in self.variables]
    self.code_lines.append(f'  return {", ".join(new_vars)}')

    # return and compile
    self.integral = utils.compile_code(
      code_scope={k: v for k, v in self.code_scope.items()},
      code_lines=self.code_lines,
      show_code=self.show_code,
      func_name=self.func_name)


register_sde_integrator('euler', Euler)


class Heun(Euler):
  def __init__(self, f, g, dt=None, name=None, show_code=False,
               var_type=None, intg_type=None, wiener_type=None,
               state_delays=None):
    if intg_type != constants.STRA_SDE:
      raise errors.IntegratorError(f'Heun method only supports Stranovich integral of SDEs, '
                                   f'but we got {intg_type} integral.')
    super(Heun, self).__init__(f=f, g=g, dt=dt, show_code=show_code, name=name,
                               var_type=var_type, intg_type=intg_type,
                               wiener_type=wiener_type, state_delays=state_delays)
    self.build()


register_sde_integrator('heun', Heun)


class Milstein(SDEIntegrator):
  def __init__(self, f, g, dt=None, name=None, show_code=False,
               var_type=None, intg_type=None, wiener_type=None,
               state_delays=None):
    super(Milstein, self).__init__(f=f, g=g, dt=dt, show_code=show_code, name=name,
                                   var_type=var_type, intg_type=intg_type,
                                   wiener_type=wiener_type, state_delays=state_delays)
    self.build()

  def build(self):
    # 2. code lines
    self.code_lines.append(f'  {constants.DT}_sqrt = {constants.DT} ** 0.5')

    # 2.1 df, dg
    df_and_dg(self.code_lines, self.variables, self.parameters)

    # 2.2 dfdt
    dfdt(self.code_lines, self.variables)

    # 2.3 dW
    noise_terms(self.code_lines, self.variables)

    # 2.3 dgdW
    # ----
    # dg * dW
    for var in self.variables:
      self.code_lines.append(f'  {var}_dgdW = {var}_dg * {var}_dW')
    self.code_lines.append('  ')

    # 2.4  df_bar = x + dfdt + math.sum(dg * dt_sqrt, axis=-1)
    all_df_bar = [f'{var}_df_bar' for var in self.variables]
    if self.wiener_type == constants.SCALAR_WIENER:
      for var in self.variables:
        self.code_lines.append(f'  {var}_df_bar = {var} + {var}_dfdt + {var}_dg * {constants.DT}_sqrt')
    else:
      for var in self.variables:
        self.code_lines.append(f'  {var}_df_bar = {var} + {var}_dfdt + math.sum('
                               f'{var}_dg * {constants.DT}_sqrt, axis=-1)')

    # 2.5  dg_bar = g(y_bar, t, *args)
    all_dg_bar = [f'{var}_dg_bar' for var in self.variables]
    self.code_lines.append(f'  {", ".join(all_dg_bar)} = g({", ".join(all_df_bar + self.parameters)})')
    self.code_lines.append('  ')

    # 2.6 dgdW2
    # ----
    # dgdW2 = 0.5 * (dg_bar - dg) * (dW * dW / dt_sqrt - dt_sqrt)
    if self.intg_type == constants.ITO_SDE:
      for var in self.variables:
        self.code_lines.append(f'  {var}_dgdW2 = 0.5 * ({var}_dg_bar - {var}_dg) * '
                               f'({var}_dW * {var}_dW / {constants.DT}_sqrt - {constants.DT}_sqrt)')
    elif self.intg_type == constants.STRA_SDE:
      for var in self.variables:
        self.code_lines.append(f'  {var}_dgdW2 = 0.5 * ({var}_dg_bar - {var}_dg) * '
                               f'{var}_dW * {var}_dW / {constants.DT}_sqrt')
    else:
      raise ValueError(f'Unknown SDE_INT type: {self.intg_type}')
    self.code_lines.append('  ')

    # 2.7 new var
    # ----
    # SCALAR_WIENER : y = x + dfdt + dgdW + dgdW2
    # VECTOR_WIENER : y = x + dfdt + math.sum(dgdW + dgdW2, axis=-1)
    if self.wiener_type == constants.SCALAR_WIENER:
      for var in self.variables:
        self.code_lines.append(f'  {var}_new = {var} + {var}_dfdt + {var}_dgdW + {var}_dgdW2')
    elif self.wiener_type == constants.VECTOR_WIENER:
      for var in self.variables:
        self.code_lines.append(f'  {var}_new = {var} + {var}_dfdt + math.sum({var}_dgdW + {var}_dgdW2, axis=-1)')
    else:
      raise ValueError(f'Unknown Wiener Process : {self.wiener_type}')
    self.code_lines.append('  ')

    # returns
    new_vars = [f'{var}_new' for var in self.variables]
    self.code_lines.append(f'  return {", ".join(new_vars)}')

    # return and compile
    self.integral = utils.compile_code(
      code_scope={k: v for k, v in self.code_scope.items()},
      code_lines=self.code_lines,
      show_code=self.show_code,
      func_name=self.func_name)


register_sde_integrator('milstein', Milstein)


class ExponentialEuler(SDEIntegrator):
  r"""First order, explicit exponential Euler method.

  For a SDE equation of the form

  .. math::

      d y=(Ay+ F(y))dt + g(y)dW(t) = f(y)dt + g(y)dW(t), \quad y(0)=y_{0}

  its schema is given by [1]_

  .. math::

      y_{n+1} & =e^{\Delta t A}(y_{n}+ g(y_n)\Delta W_{n})+\varphi(\Delta t A) F(y_{n}) \Delta t \\
       &= y_n + \Delta t \varphi(\Delta t A) f(y) + e^{\Delta t A}g(y_n)\Delta W_{n}

  where :math:`\varphi(z)=\frac{e^{z}-1}{z}`.

  References
  ----------
  .. [1] Erdoğan, Utku, and Gabriel J. Lord. "A new class of exponential integrators for stochastic
         differential equations with multiplicative noise." arXiv preprint arXiv:1608.07096 (2016).
  """

  def __init__(self, f, g, dt=None, name=None, show_code=False,
               var_type=None, intg_type=None, wiener_type=None,
               state_delays=None):
    super(ExponentialEuler, self).__init__(f=f, g=g, dt=dt, show_code=show_code, name=name,
                                           var_type=var_type, intg_type=intg_type,
                                           wiener_type=wiener_type, state_delays=state_delays)
    self.build()

  def build(self):
    # if math.get_backend_name() == 'jax':
    #   raise NotImplementedError
    # else:
    self.symbolic_build()

  def autograd_build(self):
    pass

  def symbolic_build(self):
    if self.var_type == constants.SYSTEM_VAR:
      raise errors.IntegratorError(f'Exponential Euler method do not support {self.var_type} variable type.')
    if self.intg_type != constants.ITO_SDE:
      raise errors.IntegratorError(f'Exponential Euler method only supports Ito integral, but we got {self.intg_type}.')

    if sympy is None or analysis_by_sympy is None:
      raise errors.PackageMissingError('SymPy must be installed when '
                                       'using exponential integrators.')

    # check bound method
    if hasattr(self.derivative[constants.F], '__self__'):
      self.code_lines = [f'def {self.func_name}({", ".join(["self"] + list(self.arguments))}):']

    # 1. code scope
    closure_vars = inspect.getclosurevars(self.derivative[constants.F])
    self.code_scope.update(closure_vars.nonlocals)
    self.code_scope.update(dict(closure_vars.globals))
    self.code_scope['math'] = math

    # 2. code lines
    code_lines = self.code_lines
    # code_lines = [f'def {self.func_name}({", ".join(self.arguments)}):']
    code_lines.append(f'  {constants.DT}_sqrt = {constants.DT} ** 0.5')

    # 2.1 dg
    # dg = g(x, t, *args)
    all_dg = [f'{var}_dg' for var in self.variables]
    code_lines.append(f'  {", ".join(all_dg)} = g({", ".join(self.variables + self.parameters)})')
    code_lines.append('  ')

    # 2.2 dW
    noise_terms(code_lines, self.variables)

    # 2.3 dgdW
    # ----
    # SCALAR_WIENER : dg * dW
    # VECTOR_WIENER : math.sum(dg * dW, axis=-1)

    if self.wiener_type == constants.SCALAR_WIENER:
      for var in self.variables:
        code_lines.append(f'  {var}_dgdW = {var}_dg * {var}_dW')
    else:
      for var in self.variables:
        code_lines.append(f'  {var}_dgdW = math.sum({var}_dg * {var}_dW, axis=-1)')
    code_lines.append('  ')

    # 2.4 new var
    # ----
    analysis = separate_variables(self.derivative[constants.F])
    variables_for_returns = analysis['variables_for_returns']
    expressions_for_returns = analysis['expressions_for_returns']
    for vi, (key, vars) in enumerate(variables_for_returns.items()):
      # separate variables
      sd_variables = []
      for v in vars:
        if len(v) > 1:
          raise ValueError('Cannot analyze multi-assignment code line.')
        sd_variables.append(v[0])
      expressions = expressions_for_returns[key]
      var_name = self.variables[vi]
      diff_eq = analysis_by_sympy.SingleDiffEq(var_name=var_name,
                                               variables=sd_variables,
                                               expressions=expressions,
                                               derivative_expr=key,
                                               scope=self.code_scope,
                                               func_name=self.func_name)

      f_expressions = diff_eq.get_f_expressions(substitute_vars=diff_eq.var_name)

      # code lines
      code_lines.extend([f"  {str(expr)}" for expr in f_expressions[:-1]])

      # get the linear system using sympy
      f_res = f_expressions[-1]
      df_expr = analysis_by_sympy.str2sympy(f_res.code).expr.expand()
      s_df = sympy.Symbol(f"{f_res.var_name}")
      code_lines.append(f'  {s_df.name} = {analysis_by_sympy.sympy2str(df_expr)}')
      var = sympy.Symbol(diff_eq.var_name, real=True)

      # get df part
      s_linear = sympy.Symbol(f'_{diff_eq.var_name}_linear')
      s_linear_exp = sympy.Symbol(f'_{diff_eq.var_name}_linear_exp')
      s_df_part = sympy.Symbol(f'_{diff_eq.var_name}_df_part')
      if df_expr.has(var):
        # linear
        linear = sympy.collect(df_expr, var, evaluate=False)[var]
        code_lines.append(f'  {s_linear.name} = {analysis_by_sympy.sympy2str(linear)}')
        # linear exponential
        code_lines.append(f'  {s_linear_exp.name} = math.exp({analysis_by_sympy.sympy2str(linear)} * {constants.DT})')
        # df part
        df_part = (s_linear_exp - 1) / s_linear * s_df
        code_lines.append(f'  {s_df_part.name} = {analysis_by_sympy.sympy2str(df_part)}')

      else:
        # linear exponential
        code_lines.append(f'  {s_linear_exp.name} = {constants.DT}_sqrt')
        # df part
        code_lines.append(f'  {s_df_part.name} = {s_df.name} * {constants.DT}')

      # update expression
      update = var + s_df_part

      # The actual update step
      code_lines.append(f'  {diff_eq.var_name}_new = {analysis_by_sympy.sympy2str(update)} + {var_name}_dgdW')
      code_lines.append('')

    # returns
    new_vars = [f'{var}_new' for var in self.variables]
    code_lines.append(f'  return {", ".join(new_vars)}')

    # return and compile
    self.integral = utils.compile_code(
      code_scope={k: v for k, v in self.code_scope.items()},
      code_lines=self.code_lines,
      show_code=self.show_code,
      func_name=self.func_name)

    if hasattr(self.derivative[constants.F], '__self__'):
      host = self.derivative[constants.F].__self__
      self.integral = self.integral.__get__(host, host.__class__)


register_sde_integrator('exponential_euler', ExponentialEuler)
register_sde_integrator('exp_euler', ExponentialEuler)



# -*- coding: utf-8 -*-

from typing import Union, Callable, Dict

import brainpy.math as bm
from brainpy.errors import DiffEqError
from brainpy.integrators.base import Integrator
from brainpy.integrators.constants import F, DT, unique_name
from brainpy.integrators.utils import get_args
from brainpy.tools.checking import check_dict_data

__all__ = [
  'DDEIntegrator',
]


class DDEIntegrator(Integrator):
  """Basic numerical integrator for delay differential equations (DDEs).
  """

  def __init__(
      self,
      f: Callable,
      var_type: str = None,
      dt: Union[float, int] = None,
      name: str = None,
      show_code: bool = False,
      state_delays: Dict[str, bm.TimeDelay] = None,
      neutral_delays: Dict[str, bm.NeutralDelay] = None,
  ):
    dt = bm.get_dt() if dt is None else dt
    parses = get_args(f)
    variables = parses[0]  # variable names, (before 't')
    parameters = parses[1]  # parameter names, (after 't')
    arguments = parses[2]  # function arguments

    # super initialization
    super(DDEIntegrator, self).__init__(name=name,
                                        variables=variables,
                                        parameters=parameters,
                                        arguments=arguments,
                                        dt=dt,
                                        state_delays=state_delays)

    # other settings
    self.var_type = var_type
    self.show_code = show_code

    # derivative function
    self.derivative = {F: f}
    self.f = f

    # code scope
    self.code_scope = {F: f}

    # code lines
    self.func_name = _f_names(f)
    self.code_lines = [f'def {self.func_name}({", ".join(self.arguments)}):']

    # delays
    self._neutral_delays = dict()
    if neutral_delays is not None:
      check_dict_data(neutral_delays,
                      key_type=str,
                      val_type=bm.NeutralDelay)
      for key, delay in neutral_delays.items():
        if key not in self.variables:
          raise DiffEqError(f'"{key}" is not defined in the variables: {self.variables}')
        self._neutral_delays[key] = delay
    self.register_implicit_nodes(self._neutral_delays)
    if (len(self.neutral_delays) + len(self.state_delays)) == 0:
      raise DiffEqError('There is no delay variable, it should not be '
                        'a delay differential equation, please use "brainpy.odeint()". '
                        'Or, if you forget add delay variables, please set them with '
                        '"state_delays" and "neutral_delays" arguments.')

  @property
  def neutral_delays(self):
    return self._neutral_delays

  @neutral_delays.setter
  def neutral_delays(self, value):
    raise ValueError('Cannot set "neutral_delays" by users.')

  def __call__(self, *args, **kwargs):
    assert self.integral is not None, 'Please build the integrator first.'
    # check arguments
    for i, arg in enumerate(args):
      kwargs[self.arg_names[i]] = arg

    # integral
    new_vars = self.integral(**kwargs)
    if len(self.variables) == 1:
      dict_vars = {self.variables[0]: new_vars}
    else:
      dict_vars = {k: new_vars[i] for i, k in enumerate(self.variables)}

    dt = kwargs.pop(DT, self.dt)
    # update neutral delay variables
    if len(self.neutral_delays):
      kwargs.update(dict_vars)
      new_dvars = self.f(**kwargs)
      if len(self.variables) == 1:
        new_dvars = {self.variables[0]: new_dvars}
      else:
        new_dvars = {k: new_dvars[i] for i, k in enumerate(self.variables)}
      for key, delay in self.neutral_delays.items():
        if isinstance(delay, bm.LengthDelay):
          delay.update(new_dvars[key])
        elif isinstance(delay, bm.TimeDelay):
          delay.update(kwargs['t'] + dt, new_dvars[key])
        else:
          raise ValueError('Unknown delay variable. We only supports '
                           'brainpy.math.LengthDelay, brainpy.math.TimeDelay, '
                           'brainpy.math.NeutralDelay. '
                           f'While we got {delay}')

    # update state delay variables
    for key, delay in self.state_delays.items():
      if isinstance(delay, bm.LengthDelay):
        delay.update(dict_vars[key])
      elif isinstance(delay, bm.TimeDelay):
        delay.update(kwargs['t'] + dt, dict_vars[key])
      else:
        raise ValueError('Unknown delay variable. We only supports '
                         'brainpy.math.LengthDelay, brainpy.math.TimeDelay, '
                         'brainpy.math.NeutralDelay. '
                         f'While we got {delay}')

    return new_vars


def _f_names(f):
  func_name = unique_name('dde')
  if f.__name__.isidentifier():
    func_name += '_' + f.__name__
  return func_name



# -*- coding: utf-8 -*-


import inspect
from pprint import pprint

import brainpy.math as bm
from brainpy.errors import UnsupportedError

from brainpy import errors

__all__ = [
  'get_args',
  'check_kws',
  'compile_code',
  'check_inits',
  'format_args',
]


def check_kws(parameters, keywords):
  for key, meaning in keywords.items():
    if key in parameters:
      raise errors.CodeError(f'"{key}" is a keyword for '
                             f'numerical solvers in BrainPy, denoting '
                             f'"{meaning}". Please change another name.')


def get_args(f):
  """Get the function arguments.

  >>> def f1(a, b, t, *args, c=1): pass
  >>> get_args(f1)
  (['a', 'b'], ['t', '*args', 'c'], ['a', 'b', 't', '*args', 'c=1'])

  >>> def f2(a, b, *args, c=1, **kwargs): pass
  >>> get_args(f2)
  ValueError: Do not support dict of keyword arguments: **kwargs

  >>> def f3(a, b, t, c=1, d=2): pass
  >>> get_args(f4)
  (['a', 'b'], ['t', 'c', 'd'], ['a', 'b', 't', 'c=1', 'd=2'])

  >>> def f4(a, b, t, *args): pass
  >>> get_args(f4)
  (['a', 'b'], ['t', '*args'], ['a', 'b', 't', '*args'])

  >>> scope = {}
  >>> exec(compile('def f5(a, b, t, *args): pass', '', 'exec'), scope)
  >>> get_args(scope['f5'])
  (['a', 'b'], ['t', '*args'], ['a', 'b', 't', '*args'])

  Parameters
  ----------
  f : callable
      The function.

  Returns
  -------
  args : tuple
      The variable names, the other arguments, and the original args.
  """

  # get the function arguments
  reduced_args = []
  args = []

  for name, par in inspect.signature(f).parameters.items():
    if par.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD:
      reduced_args.append(par.name)

    elif par.kind is inspect.Parameter.VAR_POSITIONAL:
      reduced_args.append(f'*{par.name}')

    elif par.kind is inspect.Parameter.KEYWORD_ONLY:
      raise errors.DiffEqError(f'In BrainPy, numerical integrators do not support KEYWORD_ONLY '
                               f'parameters, e.g., * (error in {f}).')
    elif par.kind is inspect.Parameter.POSITIONAL_ONLY:
      raise errors.DiffEqError(f'In BrainPy, numerical integrators do not support POSITIONAL_ONLY '
                               f'parameters, e.g., / (error in {f}).')
    elif par.kind is inspect.Parameter.VAR_KEYWORD:  # TODO
      raise errors.DiffEqError(f'In BrainPy, numerical integrators do not support VAR_KEYWORD '
                               f'arguments: {str(par)} (error in {f}).')
    else:
      raise errors.DiffEqError(f'Unknown argument type: {par.kind} (error in {f}).')

    args.append(str(par))

  #  variable names
  vars = []
  for a in reduced_args:
    if a == 't':
      break
    vars.append(a)
  else:
    raise ValueError('Do not find time variable "t".')
  pars = reduced_args[len(vars):]
  return vars, pars, args


def compile_code(code_lines, code_scope, func_name, show_code=False):
  code = '\n'.join(code_lines)
  if show_code:
    print(code)
    print()
    pprint(code_scope)
    print()
  exec(compile(code, '', 'exec'), code_scope)
  new_f = code_scope[func_name]
  return new_f


def check_inits(inits, variables):
  if isinstance(inits, (tuple, list)):
    assert len(inits) == len(variables), (f'Then number of variables is {len(variables)}, '
                                          f'however we only got {len(inits)} initial values.')
    inits = {v: inits[i] for i, v in enumerate(variables)}
  elif isinstance(inits, dict):
    assert len(inits) == len(variables), (f'Then number of variables is {len(variables)}, '
                                          f'however we only got {len(inits)} initial values.')
  else:
    raise UnsupportedError('Only supports dict/sequence of data for initial values. '
                           f'But we got {type(inits)}: {inits}')
  for key in list(inits.keys()):
    if key not in variables:
      raise ValueError(f'"{key}" is not defined in variables: {variables}')
    val = inits[key]
    if isinstance(val, (float, int)):
      inits[key] = bm.asarray([val], dtype=bm.float_)
  return inits


def format_args(args, kwargs, arguments):
  all_args = dict()
  for i, arg in enumerate(args):
    all_args[arguments[i]] = arg
  for key, arg in kwargs.items():
    if key in all_args:
      raise ValueError(f'{key} has been provided in *args, '
                       f'but we detect it again in **kwargs.')
    all_args[key] = arg
  return all_args



# -*- coding: utf-8 -*-


def step(vars, dt_var, A, C, code_lines, other_args):
  # steps
  for si, sval in enumerate(A):
    # k-step arguments
    k_args = []
    for v in vars:
      k_arg = f'{v}'
      for j, sv in enumerate(sval):
        if sv not in [0., '0.0', '0.', '0']:
          if sv in ['1.0', '1.', '1', 1.]:
            k_arg += f' + {dt_var} * d{v}_k{j + 1}'
          else:
            k_arg += f' + {dt_var} * d{v}_k{j + 1} * {sv}'
      if k_arg != v:
        name = f'k{si + 1}_{v}_arg'
        code_lines.append(f'  {name} = {k_arg}')
        k_args.append(name)
      else:
        k_args.append(v)

    t_arg = 't'
    if C[si] not in [0., '0.', '0']:
      if C[si] in ['1.', '1', 1.]:
        t_arg += f' + {dt_var}'
      else:
        t_arg += f' + {dt_var} * {C[si]}'
      name = f'k{si + 1}_t_arg'
      code_lines.append(f'  {name} = {t_arg}')
      k_args.append(name)
    else:
      k_args.append(t_arg)

    # k-step derivative names
    k_derivatives = [f'd{v}_k{si + 1}' for v in vars]

    # k-step code line
    code_lines.append(f'  {", ".join(k_derivatives)} = f('
                      f'{", ".join(k_args + other_args[1:])})')


def update(vars, dt_var, B, code_lines):
  return_args = []
  for v in vars:
    result = v
    for i, b1 in enumerate(B):
      if b1 not in [0., '0.', '0']:
        result += f' + d{v}_k{i + 1} * {dt_var} * {b1}'
    code_lines.append(f'  {v}_new = {result}')
    return_args.append(f'{v}_new')
  return return_args




# -*- coding: utf-8 -*-

from typing import Dict, Union

import jax.numpy as jnp
import tqdm.auto
from jax.experimental.host_callback import id_tap
from jax.tree_util import tree_map

from brainpy import math as bm
from brainpy.errors import UnsupportedError
from brainpy.nn.base import Node, Network
from brainpy.nn.utils import (check_rnn_data_time_step,
                              check_data_batch_size,
                              serialize_kwargs)
from brainpy.running.runner import Runner
from brainpy.tools.checking import check_dict_data
from brainpy.types import Tensor

__all__ = [
  'RNNRunner',
]


class RNNRunner(Runner):
  """Structural Runner for Recurrent Neural Networks.

  Parameters
  ----------
  target: Node
    The target model for simulation.
  monitors: None, list of str, tuple of str, Monitor
    Variables to monitor.
  jit: bool
    Whether we use JIT compilation to accelerate the model simulation.
  progress_bar: bool
    Whether we use progress bar to report the simulation progress.
  dyn_vars: Optional, dict
    The dynamically changed variables.
  numpy_mon_after_run : bool
    Change the monitored iterm into NumPy arrays.
  """

  def __init__(self, target: Node, jit=True, **kwargs):
    super(RNNRunner, self).__init__(target=target, **kwargs)
    assert isinstance(self.target, Node), '"target" must be an instance of brainpy.nn.Node.'

    # jit settings
    if isinstance(jit, bool):
      self.jit = {'fit': jit, 'predict': jit}
    elif isinstance(jit, dict):
      jit = {key: val for key, val in jit.items()}
      self.jit = {'fit': jit.pop('fit', True),
                  'predict': jit.pop('predict', True)}
      if len(jit):
        raise ValueError(f'Unknown jit setting for {jit.keys()}')
    else:
      raise ValueError(f'Unknown "jit" setting: {jit}')

    # function for prediction
    self._predict_func = dict()

  def __repr__(self):
    name = self.__class__.__name__
    prefix = ' ' * len(name)
    return (f'{name}(target={self.target}, \n\t'
            f'{prefix}jit={self.jit})')

  def predict(
      self,
      xs: Union[Tensor, Dict[str, Tensor]],
      forced_states: Dict[str, Tensor] = None,
      forced_feedbacks: Dict[str, Tensor] = None,
      initial_states: Union[Tensor, Dict[str, Tensor]] = None,
      initial_feedbacks: Dict[str, Tensor] = None,
      reset: bool = False,
      shared_kwargs: Dict = None,
      progress_bar: bool = True,
  ):
    """Predict a series of input data with the given target model.

    This function use the JIT compilation to accelerate the model simulation.
    Moreover, it can automatically monitor the node variables, states, inputs,
    feedbacks and its output.

    Parameters
    ----------
    xs: Tensor, dict
      The feedforward input data. It must be a 3-dimensional data
      which has the shape of `(num_sample, num_time, num_feature)`.
    forced_states: dict
      The fixed node states. Similar with ``xs``, each tensor in
      ``forced_states`` must be a tensor with the shape of
      `(num_sample, num_time, num_feature)`.
    forced_feedbacks: dict
      The fixed feedback states. Similar with ``xs``, each tensor in
      ``forced_states`` must be a tensor with the shape of
      `(num_sample, num_time, num_feature)`.
    initial_states: JaxArray, ndarray, dict
      The initial states. Each tensor in ``initial_states`` must be a
      tensor with the shape of `(num_sample, num_feature)`.
    initial_feedbacks: dict
      The initial feedbacks for the node in the network model.
      Each tensor in ``initial_feedbacks`` must be a
      tensor with the shape of `(num_sample, num_feature)`.
    reset: bool
      Whether reset the model states.
    shared_kwargs: optional, dict
      The shared arguments across different layers.
    progress_bar: bool
      Whether report the progress of the simulation using progress bar.

    Returns
    -------
    output: Tensor, dict
      The model output.
    """
    # format input data
    xs, num_step, num_batch = self._check_xs(xs)
    # set initial states
    self._set_initial_states(initial_states)
    self._set_initial_feedbacks(initial_feedbacks)
    # get forced data
    forced_states = self._check_forced_states(forced_states, num_batch, num_step)
    forced_feedbacks = self._check_forced_feedbacks(forced_feedbacks, num_batch, num_step)
    # reset the model states
    if reset:
      self.target.initialize(num_batch)
    # init monitor
    for key in self.mon.item_contents.keys():
      self.mon.item_contents[key] = []  # reshape the monitor items
    # init progress bar
    if self.progress_bar and progress_bar:
      if num_step is None:
        num_step = check_rnn_data_time_step(xs)
      self._pbar = tqdm.auto.tqdm(total=num_step)
      self._pbar.set_description(f"Predict {num_step} steps: ", refresh=True)
    # prediction
    outputs, hists = self._predict(xs=xs,
                                   forced_states=forced_states,
                                   forced_feedbacks=forced_feedbacks,
                                   shared_kwargs=shared_kwargs)
    # close the progress bar
    if self.progress_bar and progress_bar:
      self._pbar.close()
    # post-running for monitors
    for key in self.mon.item_names:
      self.mon.item_contents[key] = hists[key]
    if self.numpy_mon_after_run:
      self.mon.numpy()
    return outputs

  def _predict(
      self,
      xs: Dict[str, Tensor],
      shared_kwargs: Dict = None,
      forced_states: Dict[str, Tensor] = None,
      forced_feedbacks: Dict[str, Tensor] = None,
  ):
    """Predict the output according to the inputs.

    Parameters
    ----------
    xs: dict
      Each tensor should have the shape of `(num_time, num_batch, num_feature)`.
    forced_states: dict
      The forced state values.
    forced_feedbacks: dict
      The forced feedback output values.
    shared_kwargs: optional, dict
      The shared keyword arguments.

    Returns
    -------
    outputs, hists
      A tuple of pair of (outputs, hists).
    """
    _predict_func = self._get_predict_func(shared_kwargs)
    # rune the model
    forced_states = dict() if forced_states is None else forced_states
    forced_feedbacks = dict() if forced_feedbacks is None else forced_feedbacks
    outputs, hists = _predict_func([xs, forced_states, forced_feedbacks])
    f1 = lambda x: bm.moveaxis(x, 0, 1)
    f2 = lambda x: isinstance(x, bm.JaxArray)
    outputs = tree_map(f1, outputs, is_leaf=f2)
    hists = tree_map(f1, hists, is_leaf=f2)
    return outputs, hists

  def _get_predict_func(self, shared_kwargs: Dict = None):
    if shared_kwargs is None: shared_kwargs = dict()
    shared_kwargs_str = serialize_kwargs(shared_kwargs)
    if shared_kwargs_str not in self._predict_func:
      self._predict_func[shared_kwargs_str] = self._make_predict_func(shared_kwargs)
    return self._predict_func[shared_kwargs_str]

  def _make_predict_func(self, shared_kwargs: Dict):
    if not isinstance(shared_kwargs, dict):
      raise ValueError(f'"shared_kwargs" must be a dict, '
                       f'but got {type(shared_kwargs)}')

    def _step_func(a_input):
      xs, forced_states, forced_feedbacks = a_input
      monitors = self.mon.item_contents.keys()
      outs = self.target(xs,
                         forced_states=forced_states,
                         forced_feedbacks=forced_feedbacks,
                         monitors=monitors,
                         **shared_kwargs)
      if self.progress_bar and (self._pbar is not None):
        id_tap(lambda *args: self._pbar.update(), ())
      return outs

    if self.jit['predict']:
      dyn_vars = self.target.vars()
      dyn_vars.update(self.dyn_vars)
      f = bm.make_loop(_step_func, dyn_vars=dyn_vars.unique(), has_return=True)
      return lambda all_inputs: f(all_inputs)[1]

    else:
      def run_func(all_inputs):
        xs, forced_states, forced_feedbacks = all_inputs
        if isinstance(self.target, Network) and len(self.target.exit_nodes) > 1:
          outputs = {node.name: [] for node in self.target.exit_nodes}
          output_type = 'network'
        else:
          outputs = []
          output_type = 'node'
        monitors = {key: [] for key in self.mon.item_contents.keys()}
        num_step = check_data_batch_size(xs)
        for i in range(num_step):
          one_xs = {key: tensor[i] for key, tensor in xs.items()}
          one_forced_states = {key: tensor[i] for key, tensor in forced_states.items()}
          one_forced_feedbacks = {key: tensor[i] for key, tensor in forced_feedbacks.items()}
          output, mon = _step_func([one_xs, one_forced_states, one_forced_feedbacks])
          for key, value in mon.items():
            monitors[key].append(value)
          if output_type == 'node':
            outputs.append(output)
          else:
            for key, out in output.items():
              outputs[key].append(out)
        if output_type == 'node':
          outputs = bm.asarray(outputs)
        else:
          for key, out in outputs.items():
            outputs[key] = bm.asarray(out)
        for key, value in monitors.items():
          monitors[key] = bm.asarray(value)
        return outputs, monitors
    return run_func

  def _init_target(self, xs):  # deprecated
    # we need to initialize the node or the network
    x = dict()
    for key, tensor in xs.items():
      if not isinstance(key, str):
        raise ValueError('"xs" must a dict of (str, tensor), while we got '
                         f'({type(key)}, {type(tensor)})')
      if not isinstance(tensor, (bm.ndarray, jnp.ndarray)):
        raise ValueError('"xs" must a dict of (str, tensor), while we got '
                         f'({type(key)}, {type(tensor)})')
      x[key] = tensor[0]
    self.target.initialize(x)

  def _set_initial_states(self, initial_states):
    # initial states
    if initial_states is not None:
      if isinstance(self.target, Network):
        if not isinstance(initial_states, dict):
          raise ValueError(f'"initial_states" must be a dict when the '
                           f'target model is a brainpy.nn.Network instance. '
                           f'But we got {type(initial_states)}')
        nodes = [node.name for node in self.target.lnodes]
        for key, tensor in initial_states.items():
          if not isinstance(key, str):
            raise ValueError(f'"initial_states" must be a dict of (str, tensor). '
                             f'But got a dict of ({type(key)}, {type(tensor)})')
          if key not in nodes:
            raise ValueError(f'Node "{key}" is not defined in the target model. '
                             f'We only detect: \n{self.target.lnodes}')
          if self.target[key].state is None:
            raise ValueError(f'The target model {key} has no state. '
                             f'We cannot set its initial state.')
          self.target[key].state.value = tensor
      elif isinstance(self.target, Node):
        if self.target.state is None:
          raise ValueError(f'The target model {self.target.name} has no state. '
                           f'We cannot set its initial state.')
        if not isinstance(initial_states, (jnp.ndarray, bm.ndarray)):
          raise ValueError('"initial_states" must be a tensor, '
                           f'but we got a {type(initial_states)}')
        self.target.state.value = initial_states

  def _set_initial_feedbacks(self, initial_feedbacks):
    # initial feedback states
    if initial_feedbacks is not None:
      if isinstance(self.target, Network):
        if not isinstance(initial_feedbacks, dict):
          raise ValueError('"initial_feedbacks" must be a dict when the '
                           'target model is a brainpy.nn.Network instance. '
                           f'But we got {type(initial_feedbacks)}')
        nodes = [node.name for node in self.target.lnodes]
        for key, tensor in initial_feedbacks.items():
          if not isinstance(key, str):
            raise ValueError(f'"initial_feedbacks" must be a dict of (str, tensor). '
                             f'But got a dict of ({type(key)}, {type(tensor)})')
          if key not in nodes:
            raise ValueError(f'Node "{key}" is not defined in the target model. '
                             f'We only detect: \n{self.target.lnodes}')
          if self.target[key].fb_output is None:
            raise ValueError(f'The target model {key} has no feedback connections. '
                             f'We cannot set its initial feedback output.')
          self.target[key].fb_output.value = tensor
      elif isinstance(self.target, Node):
        raise UnsupportedError('Do not support feedback in a single instance of brainpy.nn.Node.')

  def _check_forced_states(self, forced_states, num_batch, num_step=None):
    iter_forced_states = dict()
    if forced_states is not None:
      if isinstance(self.target, Network):
        nodes = [node.name for node in self.target.lnodes]
        if not isinstance(forced_states, dict):
          raise ValueError('"forced_states" must be a dict of (str, Tensor)')
        for key, tensor in forced_states.items():
          if not isinstance(key, str):
            raise ValueError(f'"forced_states" must be a dict of (str, tensor). '
                             f'But got a dict of ({type(key)}, {type(tensor)})')
          if key not in nodes:
            raise ValueError(f'Node "{key}" is not defined in the target model. '
                             f'We only detect: \n{self.target.lnodes}')
          if not isinstance(tensor, (bm.ndarray, jnp.ndarray)):
            raise ValueError(f'"forced_states" must a dict of (str, tensor), '
                             f'while we got ({type(key)}, {type(tensor)})')
          if bm.ndim(tensor) != self.target[key].state.ndim + 1:
            raise ValueError(f'Must be a tensor with shape of (num_batch, num_time, '
                             f'{str(self.target[key].state.shape)[1:-1]}), '
                             f'but we got {tensor.shape}')
          if tensor.shape[0] != num_batch:
            raise ValueError(f'The number of the batch size ({tensor.shape[0]}) '
                             f'of the forced state of {key} does not '
                             f'match with the batch size in inputs {num_batch}.')
          if (num_step is not None) and (tensor.shape[1] != num_step):
            raise ValueError(f'The number of the time step ({tensor.shape[1]}) '
                             f'of the forced state of {key} does not '
                             f'match with the time step in inputs {num_step}.')
          if self.target[key].output_shape[1:] != tensor.shape[2:]:
            raise UnsupportedError(f'The forced state of {key} has the shape of '
                                   f'{tensor.shape}, which is not consistent with '
                                   f'its output shape {self.target[key].output_shape}. '
                                   f'Each tensor in forced state should have the shape '
                                   f'of (num_sample, num_time, num_feature) or '
                                   f'(num_sample, num_feature).')
          iter_forced_states[key] = bm.moveaxis(tensor, 0, 1)  # shape of (num_time, num_sample, num_feature)
      else:
        raise UnsupportedError('We do not support forced feedback state '
                               'for a single brainpy.nn.Node instance')
    return iter_forced_states

  def _check_forced_feedbacks(self, forced_feedbacks, num_batch, num_step):
    iter_forced_feedbacks = dict()
    if forced_feedbacks is not None:
      if isinstance(self.target, Network):
        if not isinstance(forced_feedbacks, dict):
          raise ValueError('"forced_feedbacks" must be a dict of (str, Tensor)')
        feedback_node_names = [node.name for node in self.target.feedback_nodes]
        for key, tensor in forced_feedbacks.items():
          if not isinstance(key, str):
            raise ValueError(f'"forced_feedbacks" must be a dict of (str, tensor). '
                             f'But got a dict of ({type(key)}, {type(tensor)})')
          if key not in feedback_node_names:
            raise ValueError(f'{self.target} has no feedback node {key}, '
                             f'it only has {feedback_node_names}')
          if not isinstance(tensor, (bm.ndarray, jnp.ndarray)):
            raise ValueError('"forced_feedbacks" must a dict of (str, tensor), '
                             'while we got ({type(key)}, {type(tensor)})')
          if bm.ndim(tensor) != self.target[key].fb_output.ndim + 1:
            raise ValueError(f'Must be a tensor with shape of (num_batch, num_time, '
                             f'{str(self.target[key].fb_output.shape)[1:-1]}), '
                             f'but we got {tensor.shape}')
          if tensor.shape[0] != num_batch:
            raise ValueError(f'The number of the batch size ({tensor.shape[0]}) '
                             f'of the forced feedback of {key} does not '
                             f'match with the batch size in inputs {num_batch}.')
          if tensor.shape[1] != num_step:
            raise ValueError(f'The number of the time step ({tensor.shape[1]}) '
                             f'of the forced feedback of {key} does not '
                             f'match with the time step in inputs {num_step}.')
          if self.target[key].output_shape[1:] != tensor.shape[2:]:
            raise UnsupportedError(f'The forced feedback of {key} has the shape of '
                                   f'{tensor.shape}, which is not consistent with '
                                   f'its output shape {self.target[key].output_shape}. '
                                   f'Each tensor in forced feedback should have the shape '
                                   f'of (num_sample, num_time, num_feature) or '
                                   f'(num_sample, num_feature).')
          iter_forced_feedbacks[key] = bm.moveaxis(tensor, 0, 1)  # shape of (num_time, num_sample, num_feature)
      else:
        raise UnsupportedError('We do not support forced states for '
                               'a single brainpy.nn.Node instance')
    return iter_forced_feedbacks

  def _format_xs(self, xs):
    if isinstance(xs, (bm.ndarray, jnp.ndarray)):
      if isinstance(self.target, Network):
        if len(self.target.entry_nodes) != 1:
          raise ValueError(f'The network {self.target} has {len(self.target.entry_nodes)} '
                           f'input nodes, while we only got one input data.')
        xs = {self.target.entry_nodes[0].name: xs}
      else:
        xs = {self.target.name: xs}
    if not isinstance(xs, dict):
      raise UnsupportedError(f'Unknown data type {type(xs)}, we only support '
                             f'tensor or dict with <str, tensor>')
    if len(xs) == 0:
      raise ValueError('We got no input data.')
    check_dict_data(xs, key_type=str, val_type=(bm.ndarray, jnp.ndarray))
    return xs

  def _check_xs(self, xs: Union[Dict, Tensor], move_axis=True):
    input_shapes = {}
    if isinstance(self.target, Network):
      for node in self.target.entry_nodes:
        name = self.target.entry_nodes[0].name
        input_shapes[name] = node._feedforward_shapes[name]
    else:
      name = self.target.name
      input_shapes[name] = self.target._feedforward_shapes[name]

    xs = self._format_xs(xs)
    num_times, num_batch_sizes = [], []
    for key, val in xs.items():
      if key not in input_shapes:
        raise ValueError(f'Cannot find {key} in the required inputs. Please check!')
      shape = input_shapes[key]
      if bm.ndim(val) != len(shape) + 1:
        raise ValueError(f'Each tensor in "xs" must be a tensor of shape '
                         f'(num_sample, num_time, {str(shape[1:])[1:-1]}). '
                         f'But we got {val.shape}.')
      num_times.append(val.shape[1])
      num_batch_sizes.append(val.shape[0])
    if len(set(num_times)) != 1:
      raise ValueError(f'Number of time step is different across tensors in '
                       f'the provided "xs". We got {set(num_times)}.')
    if len(set(num_batch_sizes)) != 1:
      raise ValueError(f'Number of batch size is different across tensors in '
                       f'the provided "xs". We got {set(num_batch_sizes)}.')
    num_step = num_times[0]
    num_batch = num_batch_sizes[0]
    if move_axis:
      # change shape to (num_time, num_sample, num_feature)
      xs = {k: bm.moveaxis(v, 0, 1) for k, v in xs.items()}
    return xs, num_step, num_batch




# -*- coding: utf-8 -*-

r"""This module provides explicit Runge-Kutta methods for ODEs.

Given an initial value problem specified as:

.. math::
    \frac{dy}{dt}=f(t,y),\quad y(t_{0})=y_{0}.

Let the step-size :math:`h > 0`.

Then, the general schema of explicit Runge–Kutta methods is [1]_:

.. math::
    y_{n+1}=y_{n}+h\sum _{i=1}^{s}b_{i}k_{i},

where

.. math::
    \begin{aligned}
    k_{1}&=f(t_{n},y_{n}),\\
    k_{2}&=f(t_{n}+c_{2}h,y_{n}+h(a_{21}k_{1})),\\
    k_{3}&=f(t_{n}+c_{3}h,y_{n}+h(a_{31}k_{1}+a_{32}k_{2})),\\
    &\\ \vdots \\
    k_{s}&=f(t_{n}+c_{s}h,y_{n}+h(a_{s1}k_{1}+a_{s2}k_{2}+\cdots +a_{s,s-1}k_{s-1})).
    \end{aligned}

To specify a particular method, one needs to provide the integer :math:`s` (the number
of stages), and the coefficients :math:`a_{ij}` (for :math:`1 \le j < i \le s`),
:math:`b_i` (for :math:`i = 1, 2, \cdots, s`) and :math:`c_i` (for :math:`i = 2, 3, \cdots, s`).

The matrix :math:`[a_{ij}]` is called the *Runge–Kutta matrix*, while the :math:`b_i`
and :math:`c_i` are known as the *weights* and the *nodes*. These data are usually
arranged in a mnemonic device, known as a **Butcher tableau** (named after John C. Butcher):

.. math::
    \begin{array}{c|llll}
    0 & & & & & \\
    c_{2} & a_{21} & & & & \\
    c_{3} & a_{31} & a_{32} & & & \\
    \vdots & \vdots & & \ddots & \\
    c_{s} & a_{s 1} & a_{s 2} & \cdots & a_{s, s-1} \\
    \hline & b_{1} & b_{2} & \cdots & b_{s-1} & b_{s}
    \end{array}

A Taylor series expansion shows that the Runge–Kutta method is consistent if and only if

.. math::
    \sum _{i=1}^{s}b_{i}=1.

Another popular condition for determining coefficients is:

.. math::
    \sum_{j=1}^{i-1}a_{ij}=c_{i}{\text{ for }}i=2,\ldots ,s.


More details please see references [2]_ [3]_ [4]_.

.. [1] Press, W. H., B. P. Flannery, S. A. Teukolsky, and W. T.
       Vetterling. "Section 17.1 Runge-Kutta Method." Numerical Recipes:
       The Art of Scientific Computing (2007).
.. [2] https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods
.. [3] Butcher, John Charles. Numerical methods for ordinary differential equations.
       John Wiley & Sons, 2016.
.. [4] Iserles, A., 2009. A first course in the numerical analysis of differential
       equations (No. 44). Cambridge university press.

"""

from brainpy.integrators import constants as C, utils
from brainpy.integrators.ode import common
from brainpy.integrators.ode.base import ODEIntegrator
from .generic import register_ode_integrator

__all__ = [
  'ExplicitRKIntegrator',
  'Euler',
  'MidPoint',
  'Heun2',
  'Ralston2',
  'RK2',
  'RK3',
  'Heun3',
  'Ralston3',
  'SSPRK3',
  'RK4',
  'Ralston4',
  'RK4Rule38',
]


class ExplicitRKIntegrator(ODEIntegrator):
  r"""Explicit Runge–Kutta methods for ordinary differential equation.

  For the system,

  .. math::

      \frac{d y}{d t}=f(t, y)


  Explicit Runge-Kutta methods take the form

  .. math::

      k_{i}=f\left(t_{n}+c_{i}h,y_{n}+h\sum _{j=1}^{s}a_{ij}k_{j}\right) \\
      y_{n+1}=y_{n}+h \sum_{i=1}^{s} b_{i} k_{i}

  Each method listed on this page is defined by its Butcher tableau,
  which puts the coefficients of the method in a table as follows:

  .. math::

      \begin{array}{c|cccc}
          c_{1} & a_{11} & a_{12} & \ldots & a_{1 s} \\
          c_{2} & a_{21} & a_{22} & \ldots & a_{2 s} \\
          \vdots & \vdots & \vdots & \ddots & \vdots \\
          c_{s} & a_{s 1} & a_{s 2} & \ldots & a_{s s} \\
          \hline & b_{1} & b_{2} & \ldots & b_{s}
      \end{array}

  Parameters
  ----------
  f : callable
      The derivative function.
  show_code : bool
      Whether show the formatted code.
  dt : float
      The numerical precision.
  """
  A = []  # The A matrix in the Butcher tableau.
  B = []  # The B vector in the Butcher tableau.
  C = []  # The C vector in the Butcher tableau.

  def __init__(self,
               f,
               var_type=None,
               dt=None,
               name=None,
               show_code=False,
               state_delays=None,
               neutral_delays=None):
    super(ExplicitRKIntegrator, self).__init__(f=f,
                                               var_type=var_type,
                                               dt=dt,
                                               name=name,
                                               show_code=show_code,
                                               state_delays=state_delays,
                                               neutral_delays=neutral_delays)

    # integrator keywords
    keywords = {
      C.F: 'the derivative function',
      # C.DT: 'the precision of numerical integration'
    }
    for v in self.variables:
      keywords[f'{v}_new'] = 'the intermediate value'
      for i in range(1, len(self.A) + 1):
        keywords[f'd{v}_k{i}'] = 'the intermediate value'
      for i in range(2, len(self.A) + 1):
        keywords[f'k{i}_{v}_arg'] = 'the intermediate value'
        keywords[f'k{i}_t_arg'] = 'the intermediate value'
    utils.check_kws(self.arg_names, keywords)
    self.build()

  def build(self):
    # step stage
    common.step(self.variables, C.DT,
                self.A, self.C, self.code_lines, self.parameters)
    # variable update
    return_args = common.update(self.variables, C.DT, self.B, self.code_lines)
    # returns
    self.code_lines.append(f'  return {", ".join(return_args)}')
    # compile
    self.integral = utils.compile_code(
      code_scope={k: v for k, v in self.code_scope.items()},
      code_lines=self.code_lines,
      show_code=self.show_code,
      func_name=self.func_name)


class Euler(ExplicitRKIntegrator):
  r"""The Euler method for ODEs.

  Also named as `Forward Euler method`, or `Explicit Euler` method.

  Given an ODE system,

    .. math::

        y'(t)=f(t,y(t)),\qquad y(t_{0})=y_{0},

  by using Euler method [1]_, we should choose a value :math:`h` for the
  size of every step and set :math:`t_{n}=t_{0}+nh`. Now, one step
  of the Euler method from :math:`t_{n}` to :math:`t_{n+1}=t_{n}+h` is:

  .. math::

      y_{n+1}=y_{n}+hf(t_{n},y_{n}).

  Note that the method increments a solution through an interval :math:`h`
  while using derivative information from only the beginning of the interval.
  As a result, the step's error is :math:`O(h^2)`.

  **Geometric interpretation**

  Illustration of the Euler method. The unknown curve is in blue,
  and its polygonal approximation is in red [2]_:

  .. image:: ../../../../_static/ode_Euler_method.svg
     :align: center

  **Derivation**

  There are several ways to get Euler method [2]_.

  The first is to consider the Taylor expansion of the function :math:`y`
  around :math:`t_{0}`:

  .. math::

      y(t_{0}+h)=y(t_{0})+hy'(t_{0})+{\frac {1}{2}}h^{2}y''(t_{0})+O(h^{3}).

  where :math:`y'(t_0)=f(t_0,y)`. We ignore the quadratic and higher-order
  terms, then we get Euler method. The Taylor expansion is used below to
  analyze the error committed by the Euler method, and it can be extended
  to produce Runge–Kutta methods.

  The second way is to replace the derivative with the forward finite
  difference formula:

  .. math::

      y'(t_{0})\approx {\frac {y(t_{0}+h)-y(t_{0})}{h}}.

  The third method is integrate the differential equation from :math:`t_{0}`
  to :math:`t_{0}+h` and apply the fundamental theorem of calculus to get:

  .. math::

      y(t_{0}+h)-y(t_{0})=\int _{t_{0}}^{t_{0}+h}f(t,y(t))\,\mathrm {d} t \approx hf(t_{0},y(t_{0})).

  **Note**

  Euler method is a first order numerical procedure for solving
  ODEs with a given initial value. The lack of stability
  and accuracy limits its popularity mainly to use as a
  simple introductory example of a numeric solution method.

  References
  ----------
  .. [1] W. H.; Flannery, B. P.; Teukolsky, S. A.; and Vetterling,
         W. T. Numerical Recipes in FORTRAN: The Art of Scientific
         Computing, 2nd ed. Cambridge, England: Cambridge University
         Press, p. 710, 1992.
  .. [2] https://en.wikipedia.org/wiki/Euler_method
  """
  A = [(), ]
  B = [1]
  C = [0]


register_ode_integrator('euler', Euler)


class MidPoint(ExplicitRKIntegrator):
  r"""Explicit midpoint method for ODEs.

  Also known as the `modified Euler method` [1]_.

  The midpoint method is a one-step method for numerically solving
  the differential equation given by:

  .. math::

       y'(t) = f(t, y(t)), \quad y(t_0) = y_0 .

  The formula of the explicit midpoint method is:

  .. math::

       y_{n+1} = y_n + hf\left(t_n+\frac{h}{2},y_n+\frac{h}{2}f(t_n, y_n)\right).

  Therefore, the Butcher tableau of the midpoint method is:

  .. math::

      \begin{array}{c|cc}
          0 & 0 & 0 \\
          1 / 2 & 1 / 2 & 0 \\
          \hline & 0 & 1
      \end{array}


  **Derivation**

  Compared to the slope formula of Euler method :math:`y'(t) \approx \frac{y(t+h) - y(t)}{h}`,
  the midpoint method use

  .. math::

       y'\left(t+\frac{h}{2}\right) \approx \frac{y(t+h) - y(t)}{h},

  The reason why we use this, please see the following geometric interpretation.
  Then, we get

  .. math::

       y(t+h) \approx y(t) + hf\left(t+\frac{h}{2},y\left(t+\frac{h}{2}\right)\right).

  However, we do not know :math:`y(t+h/2)`. The solution is then to use a Taylor
  series expansion exactly as the Euler method to solve:

  .. math::

      y\left(t + \frac{h}{2}\right) \approx y(t) + \frac{h}{2}y'(t)=y(t) + \frac{h}{2}f(t, y(t)),

  Finally, we can get the final step function:

  .. math::

      y(t + h) \approx y(t) + hf\left(t + \frac{h}{2}, y(t) + \frac{h}{2}f(t, y(t))\right).

  **Geometric interpretation**

  In the basic Euler's method, the tangent of the curve at :math:`(t_{n},y_{n})` is computed
  using :math:`f(t_{n},y_{n})`. The next value :math:`y_{n+1}` is found where the tangent
  intersects the vertical line :math:`t=t_{n+1}`. However, if the second derivative is only
  positive between :math:`t_{n}` and :math:`t_{n+1}`, or only negative, the curve will
  increasingly veer away from the tangent, leading to larger errors as :math:`h` increases.

  Compared with the Euler method, midpoint method use the tangent at the midpoint (upper, green
  line segment in the following figure [2]_), which would most likely give a more accurate
  approximation of the curve in that interval.

  .. image:: ../../../../_static/ode_Midpoint_method_illustration.png
     :align: center

  Although this midpoint tangent could not be accurately calculated, we can estimate midpoint
  value of :math:`y(t)` by using the original Euler's method. Finally, the improved tangent
  is used to calculate the value of :math:`y_{n+1}` from :math:`y_{n}`. This last step is
  represented by the red chord in the diagram.

  **Note**

  Note that the red chord is not exactly parallel to the green segment (the true tangent),
  due to the error in estimating the value of :math:`y(t)` at the midpoint.

  References
  ----------

  .. [1] Süli, Endre, and David F. Mayers. An Introduction to Numerical Analysis. no. 1, 2003.
  .. [2] https://en.wikipedia.org/wiki/Midpoint_method
  """
  A = [(), (0.5,)]
  B = [0, 1]
  C = [0, 0.5]


register_ode_integrator('midpoint', MidPoint)


class Heun2(ExplicitRKIntegrator):
  r"""Heun's method for ODEs.

  This method is named after Karl Heun [1]_. It is also known as
  the `explicit trapezoid rule`, `improved Euler's method`, or `modified Euler's method`.

  Given ODEs with a given initial value,

  .. math::
      y'(t) = f(t,y(t)), \qquad y(t_0)=y_0,

  the two-stage Heun's method is formulated as:

  .. math::
      \tilde{y}_{n+1} = y_n + h f(t_n,y_n)

  .. math::
      y_{n+1} = y_n + \frac{h}{2}[f(t_n, y_n) + f(t_{n+1},\tilde{y}_{n+1})],

  where :math:`h` is the step size and :math:`t_{n+1}=t_n+h`.

  Therefore, the Butcher tableau of the two-stage Heun's method is:

  .. math::
      \begin{array}{c|cc}
          0.0 & 0.0 & 0.0 \\
          1.0 & 1.0 & 0.0 \\
          \hline & 0.5 & 0.5
      \end{array}


  **Geometric interpretation**

  In the :py:func:`brainpy.integrators.ode.midpoint`, we have already known Euler
  method has big estimation error because it uses the
  line tangent to the function at the beginning of the interval :math:`t_n` as an
  estimate of the slope of the function over the interval :math:`(t_n, t_{n+1})`.

  In order to address this problem, Heun's Method considers the tangent lines to
  the solution curve at both ends of the interval (:math:`t_n` and :math:`t_{n+1}`),
  one (:math:`f(t_n, y_n)`) which *underestimates*, and one
  (:math:`f(t_{n+1},\tilde{y}_{n+1})`, approximated using Euler's Method) which
  *overestimates* the ideal vertical coordinates. The ideal point lies approximately
  halfway between the erroneous overestimation and underestimation, the average of the two slopes.

  .. image:: ../../../../_static/ode_Heun2_Method_Diagram.jpg
     :align: center

  .. math::
      \begin{aligned}
      {\text{Slope}}_{\text{left}}=&f(t_{n},y_{n}) \\
      {\text{Slope}}_{\text{right}}=&f(t_{n}+h,y_{n}+hf(t_{n},y_{n})) \\
      {\text{Slope}}_{\text{ideal}}=&{\frac {1}{2}}({\text{Slope}}_{\text{left}}+{\text{Slope}}_{\text{right}})
      \end{aligned}

  References
  ----------

  .. [1] Süli, Endre, and David F. Mayers. An Introduction to Numerical Analysis. no. 1, 2003.
  """
  A = [(), (1,)]
  B = [0.5, 0.5]
  C = [0, 1]


register_ode_integrator('heun2', Heun2)


class Ralston2(ExplicitRKIntegrator):
  r"""Ralston's method for ODEs.

  Ralston's method is a second-order method with two stages and
  a minimum local error bound.

  Given ODEs with a given initial value,

  .. math::
      y'(t) = f(t,y(t)), \qquad y(t_0)=y_0,

  the Ralston's second order method is given by

  .. math::
      y_{n+1}=y_{n}+\frac{h}{4} f\left(t_{n}, y_{n}\right)+
      \frac{3 h}{4} f\left(t_{n}+\frac{2 h}{3}, y_{n}+\frac{2 h}{3} f\left(t_{n}, y_{n}\right)\right)

  Therefore, the corresponding Butcher tableau is:

  .. math::
      \begin{array}{c|cc}
          0 & 0 & 0 \\
          2 / 3 & 2 / 3 & 0 \\
          \hline & 1 / 4 & 3 / 4
      \end{array}
  """
  A = [(), ('2/3',)]
  B = [0.25, 0.75]
  C = [0, '2/3']


register_ode_integrator('ralston2', Ralston2)


class RK2(ExplicitRKIntegrator):
  r"""Generic second order Runge-Kutta method for ODEs.

  **Derivation**

  In the :py:func:`brainpy.integrators.ode.midpoint`,
  :py:func:`brainpy.integrators.ode.heun2`, and :py:func:`brainpy.integrators.ode.ralston2`,
  we have already known first-order Euler method :py:func:`brainpy.integrators.ode.euler`
  has big estimation error.

  Here, we seek to derive a generic second order Runge-Kutta method [1]_ for the
  given ODE system with a given initial value,

  .. math::
      y'(t) = f(t,y(t)), \qquad y(t_0)=y_0,

  we want to get a generic solution:

  .. math::
      \begin{align} y_{n+1} &= y_{n} + h \left ( a_1 K_1 + a_2 K_2 \right ) \tag{1}
      \end{align}

  where :math:`a_1` and :math:`a_2` are some weights to be determined,
  and :math:`K_1` and :math:`K_2` are derivatives on the form:

  .. math::
      \begin{align}
      K_1 & = f(t_n,y_n) \qquad \text{and} \qquad K_2 = f(t_n + p_1 h,y_n + p_2 K_1 h ) \tag{2}
      \end{align}

  By substitution of (2) in (1) we get:

  .. math::
      \begin{align}
      y_{n+1} &= y_{n} + a_1 h f(t_n,y_n) + a_2 h f(t_n + p_1 h,y_n + p_2 K_1 h) \tag{3}
      \end{align}

  Now, we may find a Taylor-expansion of :math:`f(t_n + p_1 h, y_n + p_2 K_1 h )`

  .. math::
      \begin{align}
      f(t_n + p_1 h, y_n + p_2 K_1 h ) &= f + p_1 h f_t + p_2 K_1 h f_y  + \text{h.o.t.} \nonumber \\
        & = f + p_1 h f_t + p_2 h f f_y  + \text{h.o.t.}   \tag{4}
      \end{align}

  where :math:`f_t \equiv \frac{\partial f}{\partial t}` and
  :math:`f_y \equiv \frac{\partial f}{\partial y}`.

  By substitution of (4) in (3) we eliminate the implicit dependency of :math:`y_{n+1}`

  .. math::
      \begin{align}
      y_{n+1} &= y_{n} +  a_1 h f(t_n,y_n) + a_2 h \left (f + p_1 h f_t + p_2 h f f_y \right ) \nonumber \\
              &= y_{n} + (a_1 + a_2) h f + \left (a_2 p_1 f_t + a_2 p_2 f f_y \right) h^2     \tag{5}
      \end{align}

  In the next, we try to get the second order Taylor expansion of the solution:

  .. math::
      \begin{align}
       y(t_n+h) = y_n + h y' + \frac{h^2}{2} y''  + O(h^3) \tag{6}
      \end{align}

  where the second order derivative is given by

  .. math::
      \begin{align}
       y'' = \frac{d^2 y}{dt^2} = \frac{df}{dt} = \frac{\partial{f}}{\partial{t}}
       \frac{dt}{dt} + \frac{\partial{f}}{\partial{y}}  \frac{dy}{dt} = f_t + f f_y  \tag{7}
      \end{align}

  Substitution of (7) into (6) yields:

  .. math::
      \begin{align}
       y(t_n+h) = y_n + h f + \frac{h^2}{2}  \left (f_t + f f_y \right )  + O(h^3) \tag{8}
      \end{align}

  Finally, in order to approximate (8) by using (5), we get the generic second order
  Runge-Kutta method, where

  .. math::
      \begin{aligned}
      a_1 + a_2 = 1  \\
      a_2 p_1 = \frac{1}{2} \\
      a_2 p_2 = \frac{1}{2}.
      \end{aligned}

  Furthermore, let :math:`p_1=\beta`, we get

  .. math::
      \begin{aligned}
      p_1 = & \beta \\
      p_2 = & \beta \\
      a_2 = &\frac{1}{2\beta} \\
      a_1 = &1 - \frac{1}{2\beta} .
      \end{aligned}

  Therefore, the corresponding Butcher tableau is:

  .. math::

      \begin{array}{c|cc}
          0 & 0 & 0 \\
          \beta & \beta & 0 \\
          \hline & 1 - {1 \over 2 * \beta} & {1 \over 2 * \beta}
      \end{array}

  References
  ----------

  .. [1] Chapra, Steven C., and Raymond P. Canale. Numerical methods
         for engineers. Vol. 1221. New York: Mcgraw-hill, 2011.

  """

  def __init__(self,
               f,
               beta=2 / 3,
               var_type=None,
               dt=None,
               name=None,
               show_code=False,
               state_delays=None,
               neutral_delays=None):
    self.A = [(), (beta,)]
    self.B = [1 - 1 / (2 * beta), 1 / (2 * beta)]
    self.C = [0, beta]
    super(RK2, self).__init__(f=f,
                              var_type=var_type,
                              dt=dt,
                              name=name,
                              show_code=show_code,
                              state_delays=state_delays,
                              neutral_delays=neutral_delays)


register_ode_integrator('rk2', RK2)


class RK3(ExplicitRKIntegrator):
  r"""Classical third-order Runge-Kutta method for ODEs.

  For the given initial value problem :math:`y'(x) = f(t,y);\, y(t_0) = y_0`,
  the third order Runge-Kutta method is given by:

  .. math::
      y_{n+1} = y_n + 1/6 ( k_1 + 4 k_2 + k_3),

  where

  .. math::
      k_1 = h f(t_n, y_n), \\
      k_2 = h f(t_n + h / 2, y_n + k_1 / 2), \\
      k_3 = h f(t_n + h, y_n - k_1 + 2 k_2 ),

  where :math:`t_n = t_0 + n h.`

  Error term :math:`O(h^4)`,  correct up to the third order term in Taylor series expansion.

  The Taylor series expansion is :math:`y(t+h)=y(t)+\frac{k}{6}+\frac{2 k_{2}}{3}+\frac{k_{3}}{6}+O\left(h^{4}\right)`.

  The corresponding Butcher tableau is:
  
  .. math::
      \begin{array}{c|ccc}
          0 & 0 & 0 & 0 \\
          1 / 2 & 1 / 2 & 0 & 0 \\
          1 & -1 & 2 & 0 \\
          \hline & 1 / 6 & 2 / 3 & 1 / 6
      \end{array}

  """
  A = [(), (0.5,), (-1, 2)]
  B = ['1/6', '2/3', '1/6']
  C = [0, 0.5, 1]


register_ode_integrator('rk3', RK3)


class Heun3(ExplicitRKIntegrator):
  r"""Heun's third-order method for ODEs.

  It has the characteristics of:

      - method stage = 3
      - method order = 3
      - Butcher Tables:

  .. math::

      \begin{array}{c|ccc}
          0 & 0 & 0 & 0 \\
          1 / 3 & 1 / 3 & 0 & 0 \\
          2 / 3 & 0 & 2 / 3 & 0 \\
          \hline & 1 / 4 & 0 & 3 / 4
      \end{array}

  """
  A = [(), ('1/3',), (0, '2/3')]
  B = [0.25, 0, 0.75]
  C = [0, '1/3', '2/3']


register_ode_integrator('heun3', Heun3)


class Ralston3(ExplicitRKIntegrator):
  r"""Ralston's third-order method for ODEs.

  It has the characteristics of:

      - method stage = 3
      - method order = 3
      - Butcher Tables:

  .. math::
      \begin{array}{c|ccc}
          0 & 0 & 0 & 0 \\
          1 / 2 & 1 / 2 & 0 & 0 \\
          3 / 4 & 0 & 3 / 4 & 0 \\
          \hline & 2 / 9 & 1 / 3 & 4 / 9
      \end{array}

  References
  ----------

  .. [1] Ralston, Anthony (1962). "Runge-Kutta Methods with Minimum Error Bounds".
         Math. Comput. 16 (80): 431–437. doi:10.1090/S0025-5718-1962-0150954-0

  """
  A = [(), (0.5,), (0, 0.75)]
  B = ['2/9', '1/3', '4/9']
  C = [0, 0.5, 0.75]


register_ode_integrator('ralston3', Ralston3)


class SSPRK3(ExplicitRKIntegrator):
  r"""Third-order Strong Stability Preserving Runge-Kutta (SSPRK3).

  It has the characteristics of:

      - method stage = 3
      - method order = 3
      - Butcher Tables:

  .. math::
      \begin{array}{c|ccc}
          0 & 0 & 0 & 0 \\
          1 & 1 & 0 & 0 \\
          1 / 2 & 1 / 4 & 1 / 4 & 0 \\
          \hline & 1 / 6 & 1 / 6 & 2 / 3
      \end{array}

  """
  A = [(), (1,), (0.25, 0.25)]
  B = ['1/6', '1/6', '2/3']
  C = [0, 1, 0.5]


register_ode_integrator('ssprk3', SSPRK3)


class RK4(ExplicitRKIntegrator):
  r"""Classical fourth-order Runge-Kutta method for ODEs.

  For the given initial value problem of

  .. math::
      {\frac {dy}{dt}}=f(t,y),\quad y(t_{0})=y_{0}.

  The fourth-order RK method is formulated as:

  .. math::
      \begin{aligned}
      y_{n+1}&=y_{n}+{\frac {1}{6}}h\left(k_{1}+2k_{2}+2k_{3}+k_{4}\right),\\
      t_{n+1}&=t_{n}+h\\
      \end{aligned}

  for :math:`n = 0, 1, 2, 3, \cdot`, using

  .. math::
      \begin{aligned}
      k_{1}&=\ f(t_{n},y_{n}),\\
      k_{2}&=\ f\left(t_{n}+{\frac {h}{2}},y_{n}+h{\frac {k_{1}}{2}}\right),\\
      k_{3}&=\ f\left(t_{n}+{\frac {h}{2}},y_{n}+h{\frac {k_{2}}{2}}\right),\\
      k_{4}&=\ f\left(t_{n}+h,y_{n}+hk_{3}\right).
      \end{aligned}

  Here :math:`y_{n+1}` is the RK4 approximation of :math:`y(t_{n+1})`, and the next
  value (:math:`y_{n+1}`) is determined by the present value (:math:`y_{n}`) plus
  the weighted average of four increments, where each increment is the product
  of the size of the interval, :math:`h`, and an estimated slope specified by function
  :math:`f` on the right-hand side of the differential equation.

  - :math:`k_{1}` is the slope at the beginning of the interval, using :math:`y` (Euler's method);
  - :math:`k_{2}` is the slope at the midpoint of the interval, using :math:`y` and :math:`k_{1}`;
  - :math:`k_{3}` is again the slope at the midpoint, but now using :math:`y` and :math:`k_{2}`;
  - :math:`k_{4}` is the slope at the end of the interval, using :math:`y` and :math:`k_{3}`.

  The RK4 method is a fourth-order method, meaning that the local truncation error is on the order
  of (:math:`O(h^{5}`), while the total accumulated error is on the order of (:math:`O(h^{4}`).

  The corresponding Butcher tableau is:

  .. math::
      \begin{array}{c|cccc}
          0 & 0 & 0 & 0 & 0 \\
          1 / 2 & 1 / 2 & 0 & 0 & 0 \\
          1 / 2 & 0 & 1 / 2 & 0 & 0 \\
          1 & 0 & 0 & 1 & 0 \\
          \hline & 1 / 6 & 1 / 3 & 1 / 3 & 1 / 6
      \end{array}

  References
  ----------
  .. [1] Lambert, J. D. and Lambert, D. Ch. 5 in Numerical Methods for Ordinary
         Differential Systems: The Initial Value Problem. New York: Wiley, 1991.
  .. [2] Press, W. H.; Flannery, B. P.; Teukolsky, S. A.; and Vetterling, W. T.
         "Runge-Kutta Method" and "Adaptive Step Size Control for Runge-Kutta."
         §16.1 and 16.2 in Numerical Recipes in FORTRAN: The Art of Scientific
         Computing, 2nd ed. Cambridge, England: Cambridge University Press,
         pp. 704-716, 1992.
  """

  A = [(), (0.5,), (0., 0.5), (0., 0., 1)]
  B = ['1/6', '1/3', '1/3', '1/6']
  C = [0, 0.5, 0.5, 1]


register_ode_integrator('rk4', RK4)


class Ralston4(ExplicitRKIntegrator):
  r"""Ralston's fourth-order method for ODEs.

  It has the characteristics of:

      - method stage = 4
      - method order = 4
      - Butcher Tables:

  .. math::

      \begin{array}{c|cccc}
          0 & 0 & 0 & 0 & 0 \\
          .4 & .4 & 0 & 0 & 0 \\
          .45573725 & .29697761 & .15875964 & 0 & 0 \\
          1 & .21810040 & -3.05096516 & 3.83286476 & 0 \\
          \hline & .17476028 & -.55148066 & 1.20553560 & .17118478
      \end{array}

  References
  ----------

  .. [1] Ralston, Anthony (1962). "Runge-Kutta Methods with Minimum Error Bounds".
         Math. Comput. 16 (80): 431–437. doi:10.1090/S0025-5718-1962-0150954-0

  """
  A = [(), (.4,), (.29697761, .15875964), (.21810040, -3.05096516, 3.83286476)]
  B = [.17476028, -.55148066, 1.20553560, .17118478]
  C = [0, .4, .45573725, 1]


register_ode_integrator('ralston4', Ralston4)


class RK4Rule38(ExplicitRKIntegrator):
  r"""3/8-rule fourth-order method for ODEs.

  A slight variation of "the" Runge–Kutta method is also due
  to Kutta in 1901 [1]_ and is called the 3/8-rule. The primary
  advantage this method has is that almost all of the error
  coefficients are smaller than in the popular method, but it
  requires slightly more FLOPs (floating-point operations) per
  time step.


  It has the characteristics of:

      - method stage = 4
      - method order = 4
      - Butcher Tables:

  .. math::

      \begin{array}{c|cccc}
          0 & 0 & 0 & 0 & 0 \\
          1 / 3 & 1 / 3 & 0 & 0 & 0 \\
          2 / 3 & -1 / 3 & 1 & 0 & 0 \\
          1 & 1 & -1 & 1 & 0 \\
          \hline & 1 / 8 & 3 / 8 & 3 / 8 & 1 / 8
      \end{array}


  References
  ----------

  .. [1] Hairer, Ernst; Nørsett, Syvert Paul; Wanner, Gerhard (1993),
         Solving ordinary differential equations I: Nonstiff problems,
         Berlin, New York: Springer-Verlag, ISBN 978-3-540-56670-0.

  """
  A = [(), ('1/3',), ('-1/3', '1'), (1, -1, 1)]
  B = [0.125, 0.375, 0.375, 0.125]
  C = [0, '1/3', '2/3', 1]


register_ode_integrator('rk4_38rule', RK4Rule38)



# -*- coding: utf-8 -*-

"""
This module provides numerical methods for integrating Caputo fractional derivative equations.

"""

import jax.numpy as jnp
from jax.experimental.host_callback import id_tap

from brainpy import check
import brainpy.math as bm
from brainpy.errors import UnsupportedError
from brainpy.integrators.constants import DT
from brainpy.integrators.utils import check_inits, format_args
from brainpy.tools.checking import check_integer
from .base import FDEIntegrator
from .generic import register_fde_integrator

__all__ = [
  'CaputoEuler',
  'CaputoL1Schema',
]


class CaputoEuler(FDEIntegrator):
  r"""One-step Euler method for Caputo fractional differential equations.

  Given a fractional initial value problem,

  .. math::

     D_{*}^{\alpha} y(t)=f(t, y(t)), \quad y^{(k)}(0)=y_{0}^{(k)}, \quad k=0,1, \ldots,\lceil\alpha\rceil-1

  where the :math:`y_0^{(k)}` ay be arbitrary real numbers and where :math:`\alpha>0`.
  :math:`D_{*}^{\alpha}` denotes the differential operator in the sense of Caputo, defined
  by

  .. math::

     D_{*}^{\alpha} z(t)=J^{n-\alpha} D^{n} z(t)

  where :math:`n:=\lceil\alpha\rceil` is the smallest integer :math:`\geqslant \alpha`,
  Here :math:`D^n` is the usual differential operator of (integer) order :math:`n`,
  and for :math:`\mu > 0`, :math:`J^{\mu}` is the Riemann–Liouville integral operator
  of order :math:`\mu`, defined by

  .. math::

     J^{\mu} z(t)=\frac{1}{\Gamma(\mu)} \int_{0}^{t}(t-u)^{\mu-1} z(u) \mathrm{d} u

  The one-step Euler method for fractional differential equation is defined as

  .. math::

     y_{k+1} = y_0 + \frac{1}{\Gamma(\alpha)} \sum_{j=0}^{k} b_{j, k+1} f\left(t_{j}, y_{j}\right).

  where

  .. math::

     b_{j, k+1}=\frac{h^{\alpha}}{\alpha}\left((k+1-j)^{\alpha}-(k-j)^{\alpha}\right).


  Examples
  --------

  >>> import brainpy as bp
  >>>
  >>> a, b, c = 10, 28, 8 / 3
  >>> def lorenz(x, y, z, t):
  >>>   dx = a * (y - x)
  >>>   dy = x * (b - z) - y
  >>>   dz = x * y - c * z
  >>>   return dx, dy, dz
  >>>
  >>> duration = 30.
  >>> dt = 0.005
  >>> inits = [1., 0., 1.]
  >>> f = bp.fde.CaputoEuler(lorenz, alpha=0.97, num_step=int(duration / dt), inits=inits)
  >>> runner = bp.integrators.IntegratorRunner(f, monitors=list('xyz'), dt=dt, inits=inits)
  >>> runner.run(duration)
  >>>
  >>> import matplotlib.pyplot as plt
  >>> plt.plot(runner.mon.x.flatten(), runner.mon.z.flatten())
  >>> plt.show()


  Parameters
  ----------
  f : callable
    The derivative function.
  alpha: int, float, jnp.ndarray, bm.ndarray, sequence
    The fractional-order of the derivative function. Should be in the range of ``(0., 1.)``.
  num_step: int
    The total time step of the simulation.
  inits: sequence
    A sequence of the initial values for variables.
  dt: float, int
    The numerical precision.
  name: str
    The integrator name.

  References
  ----------
  .. [1] Li, Changpin, and Fanhai Zeng. "The finite difference methods for fractional
         ordinary differential equations." Numerical Functional Analysis and
         Optimization 34.2 (2013): 149-179.
  .. [2] Diethelm, Kai, Neville J. Ford, and Alan D. Freed. "Detailed error analysis
         for a fractional Adams method." Numerical algorithms 36.1 (2004): 31-52.
  """

  def __init__(self, f, alpha, num_step, inits, dt=None, name=None):
    super(CaputoEuler, self).__init__(f=f, alpha=alpha, dt=dt, name=name)

    # fractional order
    if not jnp.all(jnp.logical_and(self.alpha < 1, self.alpha > 0)):
      raise UnsupportedError(f'Only support the fractional order in (0, 1), '
                             f'but we got {self.alpha}.')

    # memory length
    check_integer(num_step, 'num_step', min_bound=1, allow_none=False)
    self.num_step = num_step

    # initial values
    self.inits = check_inits(inits, self.variables)

    # coefficients
    from scipy.special import rgamma
    rgamma_alpha = bm.asarray(rgamma(bm.as_numpy(self.alpha)))
    ranges = bm.asarray([bm.arange(num_step + 1) for _ in self.variables]).T
    coef = rgamma_alpha * bm.diff(bm.power(ranges, self.alpha), axis=0)
    self.coef = bm.flip(coef, axis=0)

    # variable states
    self.f_states = {v: bm.Variable(bm.zeros((num_step,) + self.inits[v].shape))
                     for v in self.variables}
    self.register_implicit_vars(self.f_states)
    self.idx = bm.Variable(bm.asarray([1], dtype=bm.int32))

    self.set_integral(self._integral_func)

  def _check_step(self, args, transform):
    dt, t = args
    if self.num_step * dt < t:
      raise ValueError(f'The maximum number of step is {self.num_step}, '
                       f'however, the current time {t} require a time '
                       f'step number {t / dt}.')

  def _integral_func(self, *args, **kwargs):
    # format arguments
    all_args = format_args(args, kwargs, self.arg_names)
    dt = all_args.pop(DT, self.dt)
    if check.is_checking():
      id_tap(self._check_step, (dt, all_args['t']))

    # derivative values
    devs = self.f(**all_args)
    if len(self.variables) == 1:
      if not isinstance(devs, (bm.ndarray, jnp.ndarray)):
        raise ValueError('Derivative values must be a tensor when there '
                         'is only one variable in the equation.')
      devs = {self.variables[0]: devs}
    else:
      if not isinstance(devs, (tuple, list)):
        raise ValueError('Derivative values must be a list/tuple of tensors '
                         'when there are multiple variables in the equation.')
      devs = {var: devs[i] for i, var in enumerate(self.variables)}

    # function states
    for key in self.variables:
      self.f_states[key][self.idx[0]] = devs[key]

    # integral results
    integrals = []
    idx = ((self.num_step - 1 - self.idx) + bm.arange(self.num_step)) % self.num_step
    for i, key in enumerate(self.variables):
      integral = self.inits[key] + self.coef[idx, i] @ self.f_states[key]
      integrals.append(integral * (dt ** self.alpha[i] / self.alpha[i]))
    self.idx.value = (self.idx + 1) % self.num_step

    # return integrals
    if len(self.variables) == 1:
      return integrals[0]
    else:
      return integrals


register_fde_integrator(name='CaputoEuler', integrator=CaputoEuler)


class CaputoABM(FDEIntegrator):
  """Adams-Bashforth-Moulton (ABM) Method for Caputo fractional differential equations.


  """
  pass


class CaputoL1Schema(FDEIntegrator):
  r"""The L1 scheme method for the numerical approximation of the Caputo
  fractional-order derivative equations [3]_.

  For the fractional order :math:`0<\alpha<1`, let the fractional derivative of variable
  :math:`x(t)` be

  .. math::

     \frac{d^{\alpha} x}{d t^{\alpha}}=F(x, t)

  The Caputo definition of the fractional derivative for variable :math:`x` is

  .. math::

     \frac{d^{\alpha} x}{d t^{\alpha}}=\frac{1}{\Gamma(1-\alpha)} \int_{0}^{t} \frac{x^{\prime}(u)}{(t-u)^{\alpha}} d u

  where :math:`\Gamma` is the Gamma function.

  The fractional-order derivative is capable of integrating the activity of the
  function over all past activities weighted by a function that follows a power-law.
  Using one of the numerical methods, the L1 scheme method [3]_, the numerical
  approximation of the fractional-order derivative of :math:`x` is

  .. math::

     \frac{d^{\alpha} \chi}{d t^{\alpha}} \approx \frac{(d t)^{-\alpha}}{\Gamma(2-\alpha)}\left[\sum_{k=0}^{N-1}\left[x\left(t_{k+1}\right)-
     \mathrm{x}\left(t_{k}\right)\right]\left[(N-k)^{1-\alpha}-(N-1-k)^{1-\alpha}\right]\right]

  Therefore, the numerical solution of original system is given by

  .. math::

     x\left(t_{N}\right) \approx d t^{\alpha} \Gamma(2-\alpha) F(x, t)+x\left(t_{N-1}\right)-
     \left[\sum_{k=0}^{N-2}\left[x\left(t_{k+1}\right)-x\left(t_{k}\right)\right]\left[(N-k)^{1-\alpha}-(N-1-k)^{1-\alpha}\right]\right]

  Hence, the solution of the fractional-order derivative can be described as the
  difference between the *Markov term* and the *memory trace*. The *Markov term*
  weighted by the gamma function is

  .. math::

     \text { Markov term }=d t^{\alpha} \Gamma(2-\alpha) F(x, t)+x\left(t_{N-1}\right)

  The memory trace (:math:`x`-memory trace since it is related to variable :math:`x`) is

  .. math::

     \text { Memory trace }=\sum_{k=0}^{N-2}\left[x\left(t_{k+1}\right)-x\left(t_{k}\right)\right]\left[(N-k)^{1-\alpha}-(N-(k+1))^{1-\alpha}\right]

  The memory trace integrates all the past activity and captures the long-term
  history of the system. For :math:`\alpha=1`, the memory trace is 0 for any
  time :math:`t`. When the fractional order :math:`\alpha` is decreased from 1,
  the memory trace non-linearly increases from 0, and its dynamics strongly
  depends on time. Thus, the fractional order dynamics strongly deviates
  from the first order dynamics.


  Examples
  --------

  >>> import brainpy as bp
  >>>
  >>> a, b, c = 10, 28, 8 / 3
  >>> def lorenz(x, y, z, t):
  >>>   dx = a * (y - x)
  >>>   dy = x * (b - z) - y
  >>>   dz = x * y - c * z
  >>>   return dx, dy, dz
  >>>
  >>> duration = 30.
  >>> dt = 0.005
  >>> inits = [1., 0., 1.]
  >>> f = bp.fde.CaputoL1Schema(lorenz, alpha=0.99, num_step=int(duration / dt), inits=inits)
  >>> runner = bp.integrators.IntegratorRunner(f, monitors=list('xz'), dt=dt, inits=inits)
  >>> runner.run(duration)
  >>>
  >>> import matplotlib.pyplot as plt
  >>> plt.plot(runner.mon.x.flatten(), runner.mon.z.flatten())
  >>> plt.show()


  Parameters
  ----------
  f : callable
    The derivative function.
  alpha: int, float, jnp.ndarray, bm.ndarray, sequence
    The fractional-order of the derivative function. Should be in the range of ``(0., 1.]``.
  num_step: int
    The total time step of the simulation.
  inits: sequence
    A sequence of the initial values for variables.
  dt: float, int
    The numerical precision.
  name: str
    The integrator name.

  References
  ----------
  .. [3] Oldham, K., & Spanier, J. (1974). The fractional calculus theory
         and applications of differentiation and integration to arbitrary
         order. Elsevier.
  """

  def __init__(self, f, alpha, num_step, inits, dt=None, name=None):
    super(CaputoL1Schema, self).__init__(f=f, alpha=alpha, dt=dt, name=name)

    # fractional order
    if not jnp.all(jnp.logical_and(self.alpha <= 1, self.alpha > 0)):
      raise UnsupportedError(f'Only support the fractional order in (0, 1), '
                             f'but we got {self.alpha}.')
    from scipy.special import gamma
    self.gamma_alpha = bm.asarray(gamma(bm.as_numpy(2 - self.alpha)))

    # memory length
    check_integer(num_step, 'num_step', min_bound=1, allow_none=False)
    self.num_step = num_step

    # initial values
    inits = check_inits(inits, self.variables)
    self.inits = {v: bm.Variable(inits[v]) for v in self.variables}
    self.register_implicit_vars(self.inits)

    # coefficients
    ranges = bm.asarray([bm.arange(1, num_step + 2) for _ in self.variables]).T
    coef = bm.diff(bm.power(ranges, 1 - self.alpha), axis=0)
    self.coef = bm.flip(coef, axis=0)

    # variable states
    self.diff_states = {v + "_diff": bm.Variable(bm.zeros((num_step,) + self.inits[v].shape))
                        for v in self.variables}
    self.register_implicit_vars(self.diff_states)
    self.idx = bm.Variable(bm.asarray([self.num_step - 1], dtype=bm.int32))

    # integral function
    self.set_integral(self._integral_func)

  def hists(self, var=None, numpy=True):
    if var is None:
      hists_ = {k: bm.vstack([self.inits[k], self.diff_states[k + '_diff']])
                for k in self.variables}
      hists_ = {k: bm.cumsum(v, axis=0) for k, v in hists_.items()}
      if numpy:
        hists_ = {k: v.numpy() for k, v in hists_}
      return hists_
    else:
      assert var in self.variables, (f'"{var}" is not defined in equation '
                                     f'variables: {self.variables}')
      hists_ = bm.vstack([self.inits[var], self.diff_states[var + '_diff']])
      hists_ = bm.cumsum(hists_, axis=0)
      if numpy:
        hists_ = hists_.numpy()
      return hists_

  def _check_step(self, args, transform):
    dt, t = args
    if self.num_step * dt < t:
      raise ValueError(f'The maximum number of step is {self.num_step}, '
                       f'however, the current time {t} require a time '
                       f'step number {t / dt}.')

  def _integral_func(self, *args, **kwargs):
    # format arguments
    all_args = format_args(args, kwargs, self.arg_names)
    dt = all_args.pop(DT, self.dt)
    if check.is_checking():
      id_tap(self._check_step, (dt, all_args['t']))

    # derivative values
    devs = self.f(**all_args)
    if len(self.variables) == 1:
      if not isinstance(devs, (bm.ndarray, jnp.ndarray)):
        raise ValueError('Derivative values must be a tensor when there '
                         'is only one variable in the equation.')
      devs = {self.variables[0]: devs}
    else:
      if not isinstance(devs, (tuple, list)):
        raise ValueError('Derivative values must be a list/tuple of tensors '
                         'when there are multiple variables in the equation.')
      devs = {var: devs[i] for i, var in enumerate(self.variables)}

    # integral results
    integrals = []
    idx = ((self.num_step - 1 - self.idx) + bm.arange(self.num_step)) % self.num_step
    for i, key in enumerate(self.variables):
      self.diff_states[key + '_diff'][self.idx[0]] = all_args[key] - self.inits[key]
      self.inits[key].value = all_args[key]
      markov_term = dt ** self.alpha[i] * self.gamma_alpha[i] * devs[key] + all_args[key]
      memory_trace = self.coef[idx, i] @ self.diff_states[key + '_diff']
      integral = markov_term - memory_trace
      integrals.append(integral)
    self.idx.value = (self.idx + 1) % self.num_step

    # return integrals
    if len(self.variables) == 1:
      return integrals[0]
    else:
      return integrals


register_fde_integrator(name='CaputoL1', integrator=CaputoL1Schema)
register_fde_integrator(name='CaputoL1Schema', integrator=CaputoL1Schema)
